import os
from pkg_resources import get_distribution

latest_owlfile = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), 'nidmresults', 'owl',
    "nidm-results_130.owl")

__version__ = get_distribution('nidmresults').version



"""
Objects describing the Model Parameters Estimation activity, its inputs and
outputs as specified in NIDM-Results.

Specification: http://nidm.nidash.org/specs/nidm-results.html

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""
from prov.model import Identifier
import uuid
import numpy as np
import os
from constants import *
import nibabel as nib
from generic import *
import json
import warnings


class ModelFitting(NIDMObject):

    """
    Object representing a Model fitting step: including a
    ModelParametersEstimation activity, its inputs and outputs.
    """

    def __init__(self, activity, design_matrix, data, error_model,
                 param_estimates, rms_map, mask_map, grand_mean_map,
                 machine, subjects):
        super(ModelFitting, self).__init__()
        self.activity = activity
        self.design_matrix = design_matrix
        self.data = data
        self.error_model = error_model
        self.param_estimates = param_estimates
        self.rms_map = rms_map
        self.mask_map = mask_map
        self.grand_mean_map = grand_mean_map
        self.machine = machine
        self.subjects = subjects

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # Design Matrix
        self.activity.used(self.design_matrix)
        self.add_object(self.design_matrix, nidm_version)

        if nidm_version['major'] > 1 or \
                (nidm_version['major'] == 1 and nidm_version['minor'] >= 3):
            # Machine
            self.data.wasAttributedTo(self.machine)
            self.add_object(self.machine, nidm_version)

            # Imaged subject or group(s)
            for sub in self.subjects:
                self.add_object(sub, nidm_version)
                self.data.wasAttributedTo(sub)

        # Data
        self.activity.used(self.data)
        self.add_object(self.data, nidm_version)

        # Error Model
        self.activity.used(self.error_model)
        self.add_object(self.error_model, nidm_version)

        # Parameter Estimate Maps
        for param_estimate in self.param_estimates:
            param_estimate.wasGeneratedBy(self.activity)
            self.add_object(param_estimate, nidm_version)

        # Residual Mean Squares Map
        self.rms_map.wasGeneratedBy(self.activity)
        self.add_object(self.rms_map, nidm_version)

        # Mask
        self.mask_map.wasGeneratedBy(self.activity)
        self.add_object(self.mask_map, nidm_version)

        # Grand Mean map
        self.grand_mean_map.wasGeneratedBy(self.activity)
        self.add_object(self.grand_mean_map, nidm_version)

        # Model Parameters Estimation activity
        self.add_object(self.activity, nidm_version)

        return self.p


class ImagingInstrument(NIDMObject):
    """
    Object representing a ImagingInstrument entity.
    """

    def __init__(self, machine_type):
        super(ImagingInstrument, self).__init__()
        machine_type = machine_type.lower()
        self.id = NIIRI[str(uuid.uuid4())]
        machine_term = dict(
            mri=NIF_MRI, eeg=NIF_EEG, meg=NIF_MEG, pet=NIF_PET,
            spect=NIF_SPECT)
        machine_label = dict(
            mri='MRI Scanner', eeg='EEG Machine', meg='MEG Machine',
            pet='PET Scanner', spect='SPECT Machine')
        self.type = machine_term[machine_type]
        self.prov_type = PROV['Agent']
        self.label = machine_label[machine_type]

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['label'], self.label)))

        return self.p


class Group(NIDMObject):
    """
    Object representing a ImagingInstrument entity.
    """

    def __init__(self, num_subjects, group_name, label=None, oid=None):
        super(Group, self).__init__(oid=oid)
        self.type = STATO_GROUP
        self.prov_type = PROV['Agent']
        self.group_name = group_name
        self.num_subjects = num_subjects
        if not label:
            label = "Study group population: " + group_name
        self.label = label

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_attributes((
            (PROV['type'], self.type),
            (NIDM_GROUP_NAME, self.group_name),
            (NIDM_NUMBER_OF_SUBJECTS, self.num_subjects),
            (PROV['label'], self.label)))

        return self.p


class Person(NIDMObject):
    """
    Object representing a ImagingInstrument entity.
    """

    def __init__(self):
        super(Person, self).__init__()
        self.id = NIIRI[str(uuid.uuid4())]
        self.prov_type = PROV['Agent']
        self.type = PROV['Person']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_attributes((
            (PROV['type'], self.prov_type),
            (PROV['label'], "Person")))

        return self.p


class DesignMatrix(NIDMObject):

    """
    Object representing a DesignMatrix entity.
    """

    def __init__(self, matrix, image_file, export_dir, regressors=None,
                 design_type=None, hrf_model=None, drift_model=None,
                 suffix=''):
        super(DesignMatrix, self).__init__(export_dir=export_dir)
        self.type = NIDM_DESIGN_MATRIX
        self.prov_type = PROV['Entity']
        self.matrix = matrix
        self.id = NIIRI[str(uuid.uuid4())]
        img_filename = 'DesignMatrix' + suffix + '.png'
        self.image = Image(export_dir, image_file, img_filename)
        self.regressors = regressors
        self.design_type = design_type
        self.hrf_model = hrf_model
        self.drift_model = drift_model
        self.csv_file = 'DesignMatrix' + suffix + '.csv'

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # *** Export visualisation of the design matrix
        self.add_object(self.image, nidm_version)

        # Create cvs file containing design matrix
        np.savetxt(os.path.join(self.export_dir, self.csv_file),
                   np.asarray(self.matrix), delimiter=",")

        if nidm_version['num'] in ["1.0.0", "1.1.0"]:
            csv_location = Identifier("file://./" + self.csv_file)
        else:
            csv_location = Identifier(self.csv_file)

        attributes = [(PROV['type'], self.type),
                      (PROV['label'], "Design Matrix"),
                      (NIDM_REGRESSOR_NAMES, json.dumps(self.regressors)),
                      (DCT['format'], "text/csv"),
                      (NFO['fileName'], self.csv_file),
                      (DC['description'], self.image.id),
                      (PROV['location'], csv_location)]

        if self.hrf_model is not None:
            if nidm_version['num'] in ("1.0.0", "1.1.0"):
                if self.design_type is not None:
                    attributes.append(
                        (NIDM_HAS_FMRI_DESIGN, self.design_type))
                else:
                    warnings.warn("Design type is missing")

            # hrf model
            attributes.append((NIDM_HAS_HRF_BASIS, self.hrf_model))
            # drift model
            self.add_object(self.drift_model, nidm_version)
            attributes.append((NIDM_HAS_DRIFT_MODEL, self.drift_model.id))

        # Create "design matrix" entity
        self.add_attributes(attributes)

        return self.p


class DriftModel(NIDMObject):

    """
    Object representing a DriftModel entity.
    """

    def __init__(self, drift_type, parameter):
        super(DriftModel, self).__init__()
        self.drift_type = drift_type
        self.id = NIIRI[str(uuid.uuid4())]
        self.parameter = parameter
        self.type = drift_type
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        attributes = [(PROV['type'], self.drift_type)]

        if self.drift_type == FSL_GAUSSIAN_RUNNING_LINE_DRIFT_MODEL:
            attributes.append(
                (PROV['label'], "FSL's Gaussian Running Line Drift Model"))
            attributes.append((FSL_DRIFT_CUTOFF_PERIOD, self.parameter))

        # Create "drift model" entity
        self.add_attributes(attributes)

        return self.p


class Data(NIDMObject):

    """
    Object representing a Data entity.
    """

    def __init__(self, grand_mean_scaling, target, mri_protocol=None,
                 label=None, group_or_sub=None, oid=None):
        super(Data, self).__init__(oid=oid)
        self.grand_mean_sc = grand_mean_scaling
        self.target_intensity = target
        self.type = NIDM_DATA
        self.prov_type = PROV['Entity']
        self.mri_protocol = mri_protocol
        if self.mri_protocol == "fmri":
            self.mri_protocol = NLX_FMRI_PROTOCOL
        if label is None:
            label = "Data"
        self.label = label
        self.group_or_sub = group_or_sub

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        if nidm_version['major'] < 1 or \
                (nidm_version['major'] == 1 and nidm_version['minor'] < 3):
            self.type = NIDM_DATA_SCALING
    # Create "Data" entity
        # FIXME: grand mean scaling?
        # FIXME: medianIntensity
        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['type'], PROV['Collection']),
            (PROV['label'], self.label),
            (NIDM_GRAND_MEAN_SCALING, self.grand_mean_sc),
            (NIDM_TARGET_INTENSITY, self.target_intensity)))

        if nidm_version['major'] > 1 or \
                (nidm_version['major'] == 1 and nidm_version['minor'] > 2):
            if self.mri_protocol is not None:
                self.add_attributes(
                    [(NIDM_HAS_MRI_PROTOCOL, self.mri_protocol)])

        return self.p


class ErrorModel(NIDMObject):

    """
    Object representing an ErrorModel entity.
    """

    def __init__(self, error_distribution, variance_homo, variance_spatial,
                 dependance, dependance_spatial):
        super(ErrorModel, self).__init__()
        self.error_distribution = error_distribution
        self.variance_homo = variance_homo
        self.variance_spatial = variance_spatial
        self.dependance = dependance
        self.dependance_spatial = dependance_spatial
        self.id = NIIRI[str(uuid.uuid4())]
        self.type = NIDM_ERROR_MODEL
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        atts = (
            (PROV['type'], NIDM_ERROR_MODEL),
            (NIDM_HAS_ERROR_DISTRIBUTION, self.error_distribution),
            (NIDM_ERROR_VARIANCE_HOMOGENEOUS, self.variance_homo),
            (NIDM_VARIANCE_SPATIAL_MODEL, self.variance_spatial),
            (NIDM_HAS_ERROR_DEPENDENCE, self.dependance))

        # If the error covariance is independent then there is no associated
        # spatial model
        if self.dependance_spatial is not None:
            atts = atts + (
                ((NIDM_DEPENDENCE_SPATIAL_MODEL, self.dependance_spatial),))

        # Create "Error Model" entity
        self.add_attributes(atts)

        return self.p


class ModelParametersEstimation(NIDMObject):

    """
    Object representing an ModelParametersEstimation activity.
    """

    def __init__(self, estimation_method, software_id, data=None, label=None,
                 oid=None):
        super(ModelParametersEstimation, self).__init__(oid=oid)
        self.estimation_method = estimation_method
        self.software_id = software_id
        self.type = NIDM_MODEL_PARAMETERS_ESTIMATION
        self.prov_type = PROV['Activity']
        # currenlty only used for reading
        self.data = data
        if label is None:
            label = "Model Parameters Estimation"
        self.label = label

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # Create "Model Parameter estimation" activity
        self.add_attributes((
            (PROV['type'], self.type),
            (NIDM_WITH_ESTIMATION_METHOD, self.estimation_method),
            (PROV['label'], self.label)))

        return self.p


class ParameterEstimateMap(NIDMObject):

    """
    Object representing an ParameterEstimateMap entity.
    """

    def __init__(self, pe_file, pe_num, coord_space, filename=None, sha=None,
                 label=None, suffix='', model_param_estimation=None, oid=None):
        super(ParameterEstimateMap, self).__init__(oid=oid)
        # Column index in the corresponding design matrix
        self.num = pe_num
        self.coord_space = coord_space
        self.file = NIDMFile(self.id, pe_file, new_filename=filename, sha=sha)
        self.type = NIDM_PARAMETER_ESTIMATE_MAP
        self.prov_type = PROV['Entity']
        if label is None:
            label = "Parameter estimate " + str(self.num)
        self.label = label
        # Only used for reading (so far)
        self.model_param_estimation = model_param_estimation

    # Generate prov for contrast map
    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_object(self.coord_space, nidm_version)
        self.add_object(self.file, nidm_version)

        # Parameter estimate entity
        self.add_attributes((
            (PROV['type'], self.type),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
            (PROV['label'], self.label)))

        return self.p


class ResidualMeanSquares(NIDMObject):

    """
    Object representing an ResidualMeanSquares entity.
    """

    def __init__(self, export_dir, residual_file, coord_space,
                 temporary=False, suffix=''):
        super(ResidualMeanSquares, self).__init__(export_dir)
        self.coord_space = coord_space
        self.id = NIIRI[str(uuid.uuid4())]
        filename = 'ResidualMeanSquares' + suffix + '.nii.gz'
        self.file = NIDMFile(self.id, residual_file, filename, export_dir,
                             temporary=temporary)
        self.type = NIDM_RESIDUAL_MEAN_SQUARES_MAP
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # Create coordinate space export
        self.add_object(self.coord_space, nidm_version)

        # Create "residuals map" entity
        self.add_object(self.file, nidm_version)

        self.add_attributes((
            (PROV['type'], self.type,),
            (PROV['label'], "Residual Mean Squares Map"),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id)))

        return self.p


class MaskMap(NIDMObject):

    """
    Object representing an MaskMap entity.
    """

    def __init__(self, export_dir, mask_file, coord_space, user_defined,
                 suffix=''):
        super(MaskMap, self).__init__(export_dir)
        self.coord_space = coord_space
        self.id = NIIRI[str(uuid.uuid4())]
        filename = 'Mask' + suffix + '.nii.gz'
        self.file = NIDMFile(self.id, mask_file, filename, export_dir)
        self.user_defined = user_defined
        self.type = NIDM_MASK_MAP
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # Create coordinate space export
        self.add_object(self.coord_space, nidm_version)

        # Create "Mask map" entity
        self.add_object(self.file, nidm_version)

        self.add_attributes((
            (PROV['type'], self.type,),
            (PROV['label'], "Mask"),
            (NIDM_IS_USER_DEFINED, self.user_defined),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id))
        )

        return self.p


class GrandMeanMap(NIDMObject):

    """
    Object representing an GrandMeanMap entity.
    """

    def __init__(self, org_file, mask_file, coord_space, export_dir,
                 suffix=''):
        super(GrandMeanMap, self).__init__(export_dir)
        self.id = NIIRI[str(uuid.uuid4())]
        filename = 'GrandMean' + suffix + '.nii.gz'
        self.file = NIDMFile(self.id, org_file, filename, export_dir)
        self.mask_file = mask_file  # needed to compute masked median
        self.coord_space = coord_space
        self.type = NIDM_GRAND_MEAN_MAP
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # Coordinate space entity
        self.add_object(self.coord_space, nidm_version)

        # Grand Mean Map entity
        self.add_object(self.file, nidm_version)

        grand_mean_file = self.file.path
        grand_mean_img = nib.load(grand_mean_file)
        grand_mean_data = grand_mean_img.get_data()
        grand_mean_data = np.ndarray.flatten(grand_mean_data)

        mask_img = nib.load(self.mask_file)
        mask_data = mask_img.get_data()
        mask_data = np.ndarray.flatten(mask_data)

        grand_mean_data_in_mask = grand_mean_data[mask_data > 0]
        masked_median = np.median(
            np.array(grand_mean_data_in_mask, dtype=float))

        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['label'], "Grand Mean Map"),
            (NIDM_MASKED_MEDIAN, masked_median),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id))
        )

        return self.p



"""
Export neuroimaging results created by neuroimaging software packages
(FSL, AFNI, ...) following NIDM-Results specification.

Specification: http://nidm.nidash.org/specs/nidm-results.html

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""

from nidmresults.objects.constants import *
from nidmresults.objects.modelfitting import *
from nidmresults.objects.contrast import *
from nidmresults.objects.inference import *

from rdflib.plugins.parsers.notation3 import BadSyntax
import zipfile
import csv


class Graph():
    """
    Generic class to read a NIDM-result archive and create a python object.
    """

    def __init__(self, nidm_zip=None, rdf_file=None, format="turtle"):
        with zipfile.ZipFile(nidm_zip) as z:
            rdf_data = z.read('nidm.ttl')

        self.study_name = os.path.basename(nidm_zip).replace(".nidm.zip", "")

        self.rdf_data = rdf_data

        self.format = format
        self.graph = self.parse()

        self.objects = dict()

    def parse(self):
        g = rdflib.Graph()
        try:
            g.parse(data=self.rdf_data, format=self.format)
        except BadSyntax:
            raise self.ParseException(
                "RDFLib was unable to parse the RDF file.")
        return g

    def get_statistic_maps(self):
        """
        Read a NIDM-Results document and return a list of Statistic Maps.
        """

        query = """
        prefix prov: <http://www.w3.org/ns/prov#>
        prefix nidm_contrastName: <http://purl.org/nidash/nidm#NIDM_0000085>
        prefix nidm_StatisticMap: <http://purl.org/nidash/nidm#NIDM_0000076>
        prefix nidm_statisticType: <http://purl.org/nidash/nidm#NIDM_0000123>
        prefix nidm_errorDegreesOfFreedom: <http://purl.org/nidash/nidm#NIDM_0\
000093>

        SELECT ?label ?contrastName ?statType ?statFile ?dof WHERE {
         ?sid a nidm_StatisticMap: ;
              nidm_contrastName: ?contrastName ;
              nidm_statisticType: ?statType ;
              rdfs:label ?label ;
              nidm_errorDegreesOfFreedom: ?dof ;
              prov:atLocation ?statFile .
        }
        """
        sd = self.graph.query(query)

        stat_maps = list()
        if sd:
            for label, contrast_name, stat_type, stat_file, dof in sd:
                contrast_num = None
                coord_space = None
                export_dir = None
                stat_maps.append(StatisticMap(
                    stat_file, stat_type, contrast_num, contrast_name, dof,
                    coord_space, export_dir, label))
        self.stat_maps = stat_maps
        return stat_maps

    def get_peaks(self, contrast_name=None):
        """
        Read a NIDM-Results document and return a list of Peaks.
        """

        if contrast_name is not None:
            query_extension = """
?cluster a significant_cluster: .
?peak prov:wasDerivedFrom ?cluster .
?cluster prov:wasDerivedFrom/prov:wasGeneratedBy/prov:used/prov:\
wasGeneratedBy ?conest .
?conmap prov:wasGeneratedBy ?conest .
?conmap a nidm_ContrastMap: .
?conmap nidm_contrastName: ?conname .
            """
        else:
            query_extension = ""

        query = """
prefix prov: <http://www.w3.org/ns/prov#>
prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
prefix nidm_ExcursionSetMap: <http://purl.org/nidash/nidm#NIDM_0000025>
prefix nidm_pValueUncorrected: <http://purl.org/nidash/nidm#NIDM_0000116>
prefix nidm_equivalentZStatistic: <http://purl.org/nidash/nidm#NIDM_0000092>
prefix nidm_coordinateVector: <http://purl.org/nidash/nidm#NIDM_0000086>
prefix nidm_Coordinate: <http://purl.org/nidash/nidm#NIDM_0000015>
prefix nidm_contrastName: <http://purl.org/nidash/nidm#NIDM_0000085>

SELECT DISTINCT ?coord_label ?coord_vector ?z ?peak_label ?p_unc
?peak_id ?exc_set_id WHERE {
    ?coord a nidm_Coordinate: ;
        rdfs:label ?coord_label ;
        nidm_coordinateVector: ?coord_vector .
    ?peak_id prov:atLocation ?coord ;
        rdfs:label ?peak_label ;
        nidm_equivalentZStatistic: ?z ;
        nidm_pValueUncorrected: ?p_unc ;
        prov:wasDerivedFrom/prov:wasDerivedFrom ?exc_set_id .
    """ + query_extension + """
}
ORDER BY ?peak_label
        """
        sd = self.graph.query(query)

        peaks = dict()
        if sd:
            for coord_label, coord_vector, z, peak_label, p_unc, peak_id, \
                    exc_set_id in sd:
                peak = Peak(float(z), coord_vector=coord_vector,
                            p_unc=float(p_unc), label=peak_label,
                            coord_label=coord_label,
                            oid=peak_id, exc_set_id=exc_set_id)
                peaks[peak_id] = (peak)

        self.objects.update(peaks)
        self.peaks = peaks
        return peaks

    def peak_query(self):
        query = """
        prefix prov: <http://www.w3.org/ns/prov#>
        prefix spm: <http://purl.org/nidash/spm#>
        prefix nidm: <http://purl.org/nidash/nidm#>
        prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>

        prefix peak: <http://purl.org/nidash/nidm#NIDM_0000062>
        prefix significant_cluster: <http://purl.org/nidash/nidm#NIDM_0000070>
        prefix coordinate: <http://purl.org/nidash/nidm#NIDM_0000086>
        prefix equivalent_zstatistic: <http://purl.org/nidash/\
nidm#NIDM_0000092>
        prefix pvalue_fwer: <http://purl.org/nidash/nidm#NIDM_0000115>
        prefix pvalue_uncorrected: <http://purl.org/nidash/nidm#NIDM_0000116>
        prefix nidm_ContrastMap: <http://purl.org/nidash/nidm#NIDM_0000002>
        prefix nidm_contrastName: <http://purl.org/nidash/nidm#NIDM_0000085>

        ####################
        #Peaks Data Query
        ####################
        SELECT ?peak ?xyz ?cluster ?zstat ?pvalfwer ?conname
        WHERE
        { ?peak a peak: .
          ?cluster a significant_cluster: .
          ?peak prov:wasDerivedFrom ?cluster .
          ?peak prov:atLocation ?coordinate .
          ?coordinate coordinate: ?xyz .
          ?peak equivalent_zstatistic: ?zstat .
          OPTIONAL { ?peak pvalue_fwer: ?pvalfwer }.
          ?cluster prov:wasDerivedFrom/prov:wasGeneratedBy/prov:used/prov:\
wasGeneratedBy ?conest .
          ?conmap prov:wasGeneratedBy ?conest .
          ?conmap a nidm_ContrastMap: .
          ?conmap nidm_contrastName: ?conname .
        }
        ORDER BY ?cluster ?peak
        """
        sd = self.graph.query(query)

        peaks = list()
        if sd:
            for peak_id, xyz, cluster, zstat, pfwer, conname in sd:
                local_peak_id = None
                peak = Peak(None, local_peak_id, zstat, 1, cluster_id=cluster,
                            coord_vector=xyz, p_fwer=pfwer, oid=peak_id)
                peaks.append(peak)
        return peaks_df

    def get_coord_spaces(self, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_CoordinateSpace: <http://purl.org/nidash/nidm#NIDM_0000016>
prefix nidm_voxelToWorldMapping: <http://purl.org/nidash/nidm#NIDM_0000132>
prefix nidm_voxelUnits: <http://purl.org/nidash/nidm#NIDM_0000133>
prefix nidm_voxelSize: <http://purl.org/nidash/nidm#NIDM_0000131>
prefix nidm_inWorldCoordinateSystem: <http://purl.org/nidash/nidm#NIDM_0000105>
prefix nidm_MNICoordinateSystem: <http://purl.org/nidash/nidm#NIDM_0000051>
prefix nidm_numberOfDimensions: <http://purl.org/nidash/nidm#NIDM_0000112>
prefix nidm_dimensionsInVoxels: <http://purl.org/nidash/nidm#NIDM_0000090>


SELECT ?oid ?label ?vox_to_world ?units ?vox_size ?coordinate_system ?numdim
?dimensions
        WHERE
        {
    """ + oid_var + """ a nidm_CoordinateSpace: ;
    rdfs:label ?label ;
    nidm_voxelToWorldMapping: ?vox_to_world ;
    nidm_voxelUnits: ?units ;
    nidm_voxelSize: ?vox_size ;
    nidm_inWorldCoordinateSystem: ?coordinate_system ;
    nidm_numberOfDimensions: ?numdim ;
    nidm_dimensionsInVoxels: ?dimensions .
    }
        """
        sd = self.graph.query(query)

        coord_spaces = dict()
        if sd:
            for row in sd:
                args = row.asdict()
                if oid is not None:
                    args['oid'] = oid
                coord_spaces[args['oid']] = CoordinateSpace(**args)

        if oid is not None:
            return coord_spaces[oid]
        else:
            return coord_spaces

    def get_groups(self, oid=None):
        """
        Read a NIDM-Results document and return a dict of Groups.
        """
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix obo_studygrouppopulation: <http://purl.obolibrary.org/obo/STATO_0000193>
prefix nidm_groupName: <http://purl.org/nidash/nidm#NIDM_0000170>
prefix nidm_numberOfSubjects: <http://purl.org/nidash/nidm#NIDM_0000171>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a obo_studygrouppopulation: ;
        rdfs:label ?label ;
        nidm_groupName: ?group_name ;
        nidm_numberOfSubjects: ?num_subjects .
}
        """
        objects = dict()

        arg_list = self.run_query_and_get_args(query, oid)
        for args in arg_list:
            group = Group(**args)
            objects[args['oid']] = group

        self.objects.update(objects)
        if oid is not None:
            return objects[oid]
        else:
            return objects

    def get_datas(self, oid=None):
        """
        Read a NIDM-Results document and return a dict of Data.
        """
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_Data: <http://purl.org/nidash/nidm#NIDM_0000169>
prefix obo_studygrouppopulation: <http://purl.obolibrary.org/obo/STATO_0000193>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_Data: ;
        rdfs:label ?label ;
        prov:wasAttributedTo ?group_id .
    ?group_id a obo_studygrouppopulation: .
}
        """
        objects = dict()

        arg_list = self.run_query_and_get_args(query, oid)
        for args in arg_list:
            # FIXME: load from file
            args["grand_mean_scaling"] = None
            args["target"] = None
            args["mri_protocol"] = None
            args["group_or_sub"] = args["group"]
            args.pop("group", None)
            # (self, grand_mean_scaling, target, mri_protocol=None, oid=None)
            objects[args['oid']] = Data(**args)

        self.objects.update(objects)

        if oid is not None:
            return objects[oid]
        else:
            return objects

    def get_model_param_estimations(self, oid=None):
        """
        Read a NIDM-Results document and return a dict of Model Parameter
        Estimations.
        """
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_ModelParameterEstimation: <http://purl.org/nidash/nidm#NIDM_000005\
6>
prefix nidm_ParameterEstimateMap: <http://purl.org/nidash/nidm#NIDM_0000061>
prefix nidm_Data: <http://purl.org/nidash/nidm#NIDM_0000169>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_ModelParameterEstimation: ;
        rdfs:label ?label ;
        prov:used ?data_id .
    ?data_id a nidm_Data: .
}
        """
        objects = dict()

        arg_list = self.run_query_and_get_args(query, oid)
        for args in arg_list:
            args['estimation_method'] = None
            args['software_id'] = None
            # FIXME: read from file
            objects[args['oid']] = ModelParametersEstimation(**args)
            # (self, estimation_method, software_id):

        self.objects.update(objects)
        if oid is not None:
            return objects[oid]
        else:
            return objects

    def get_param_estimates(self, oid=None):
        """
        Read a NIDM-Results document and return a dict of Parameter Estimate
        Maps.
        """
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_ParameterEstimateMap: <http://purl.org/nidash/nidm#NIDM_0000061>
prefix nidm_inCoordinateSpace: <http://purl.org/nidash/nidm#NIDM_0000104>
prefix nidm_ModelParameterEstimation: <http://purl.org/nidash/nidm#NIDM_000005\
6>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_ParameterEstimateMap: ;
        rdfs:label ?label ;
        nfo:fileName ?filename ;
        nidm_inCoordinateSpace: ?coord_space_id ;
        crypto:sha512 ?sha ;
        dct:format ?format ;
        prov:wasGeneratedBy ?model_param_estimation_id .
    ?model_param_estimation_id a nidm_ModelParameterEstimation: .
}
        """
        objects = dict()

        arg_list = self.run_query_and_get_args(query, oid)

        pe_num = 1
        for args in arg_list:
            args.pop("format", None)
            args["pe_num"] = pe_num
            args["pe_file"] = None
            pe_num = pe_num + 1
            objects[args['oid']] = ParameterEstimateMap(**args)

        self.objects.update(objects)
        if oid is not None:
            return objects[oid]
        else:
            return objects

    def get_contrast_estimations(self, oid=None):
        """
        Read a NIDM-Results document and return a dict of Contrast Estimations.
        """
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_ContrastEstimation: <http://purl.org/nidash/nidm#NIDM_0000001>
prefix nidm_ParameterEstimateMap: <http://purl.org/nidash/nidm#NIDM_0000061>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_ContrastEstimation: ;
        rdfs:label ?label ;
        prov:used ?param_estimate_id .
    ?param_estimate_id a nidm_ParameterEstimateMap: .
}
        """
        objects = dict()

        arg_list = self.run_query_and_get_args(query, oid)

        con_num = 1
        for args in arg_list:
            # Assign random contrast number
            args['contrast_num'] = con_num
            con_num = con_num + 1
            # FIXME: deal with more than one param_estimate as input
            objects[args['oid']] = ContrastEstimation(**args)

        self.objects.update(objects)
        if oid is not None:
            return objects[oid]
        else:
            return objects

    def get_stat_maps(self, oid=None):
        """
        Read a NIDM-Results document and return a dict of Statistic Maps.
        """
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_StatisticMap: <http://purl.org/nidash/nidm#NIDM_0000076>
prefix nidm_statisticType: <http://purl.org/nidash/nidm#NIDM_0000123>
prefix nidm_contrastName: <http://purl.org/nidash/nidm#NIDM_0000085>
prefix nidm_effectDegreesOfFreedom: <http://purl.org/nidash/nidm#NIDM_0000091>
prefix nidm_inCoordinateSpace: <http://purl.org/nidash/nidm#NIDM_0000104>
prefix nidm_errorDegreesOfFreedom: <http://purl.org/nidash/nidm#NIDM_0000093>
prefix obo_tstatistic: <http://purl.obolibrary.org/obo/STATO_0000176>
prefix nidm_ContrastEstimation: <http://purl.org/nidash/nidm#NIDM_0000001>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_StatisticMap: ;
        rdfs:label ?label ;
        prov:atLocation ?location ;
        nidm_statisticType: ?stat_type ;
        nfo:fileName ?filename ;
        dct:format ?format ;
        nidm_contrastName: ?contrast_name ;
        nidm_errorDegreesOfFreedom: ?dof ;
        nidm_effectDegreesOfFreedom: ?effdof ;
        nidm_inCoordinateSpace: ?coord_space_id ;
        crypto:sha512 ?sha ;
        prov:wasGeneratedBy ?contrast_estimation_id .
    ?contrast_estimation_id a nidm_ContrastEstimation: .
}
        """
        stat_maps = dict()

        arg_list = self.run_query_and_get_args(query, oid)
        for args in arg_list:
            # FIXME: will have to be set by default but that will change
            # position of arguments (check for compatibility)
            args['contrast_num'] = None
            stat_maps[args['oid']] = StatisticMap(**args)

        self.objects.update(stat_maps)
        if oid is not None:
            return stat_maps[oid]
        else:
            return stat_maps

    def run_query_and_get_args(self, query, oid):
        sd = self.graph.query(query)
        all_args = list()
        if sd:
            for row in sd:
                args = row.asdict()

                for key, value in args.iteritems():
                    if key.endswith("_id"):
                        object_name = key.replace("_id", "")
                        method_name = "get_" + object_name + "s"
                        method = getattr(self, method_name)

                        loaded_object = method(value)
                        args[object_name] = loaded_object
                        args.pop(key, None)

                if oid is not None:
                    args['oid'] = oid

                all_args.append(args)
        else:
            raise Exception('No results for query:\n' + query)

        return all_args

    def get_inferences(self, oid=None):
        """
        Read a NIDM-Results document and return a dict of Inference activities.
        """
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_Inference: <http://purl.org/nidash/nidm#NIDM_0000049>
prefix nidm_hasAltHypothesis: <http://purl.org/nidash/nidm#NIDM_0000097>
prefix nidm_OneTailedTest: <http://purl.org/nidash/nidm#NIDM_0000060>
prefix nidm_StatisticMap: <http://purl.org/nidash/nidm#NIDM_0000076>

SELECT DISTINCT ?oid ?label ?tail ?stat_map_id WHERE {
    """ + oid_var + """ a nidm_Inference: ;
        rdfs:label ?label ;
        nidm_hasAltHypothesis: ?tail ;
        prov:used ?stat_map_id .
    ?stat_map_id a nidm_StatisticMap: .
}
        """
        sd = self.graph.query(query)

        inferences = dict()
        if sd:
            for row in sd:
                stat_map = self.get_stat_maps(row.stat_map_id)
                args = row.asdict()
                args['stat_map'] = stat_map
                args['contrast_num'] = None
                args.pop("stat_map_id", None)
                if oid is not None:
                    args['oid'] = oid

                inferences[args['oid']] = InferenceActivity(**args)

        self.objects.update(inferences)
        if oid is not None:
            return inferences[oid]
        else:
            return inferences

    def get_excursion_set_maps(self):
        """
        Read a NIDM-Results document and return a dict of ExcursionSet.
        """

        # We need to add optional
    #         nidm_hasClusterLabelsMap: ?cluster_label_map_id ;
    # nidm_hasMaximumIntensityProjection: ?mip_id ;
        # nidm_numberOfSignificantClusters: ?num_signif_vox ;
        # nidm_pValue: ?p_value ;
        # visu
        query = """
prefix prov: <http://www.w3.org/ns/prov#>
prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>

prefix nidm_ExcursionSetMap: <http://purl.org/nidash/nidm#NIDM_0000025>
prefix nidm_hasClusterLabelsMap: <http://purl.org/nidash/nidm#NIDM_0000098>
prefix nidm_hasMaximumIntensityProjection: <http://purl.org/nidash/nidm#NIDM_0\
000138>
prefix nidm_inCoordinateSpace: <http://purl.org/nidash/nidm#NIDM_0000104>
prefix nidm_numberOfSignificantClusters: <http://purl.org/nidash/nidm#NIDM_000\
0111>
prefix nidm_pValue: <http://purl.org/nidash/nidm#NIDM_0000114>

SELECT DISTINCT ?oid ?label ?location ?format ?filname ?cluster_label_map_id
?mip_id ?coord_space_id ?sha ?num_signif_vox ?p_value ?inference_id WHERE {

?oid a nidm_ExcursionSetMap: ;
    prov:atLocation ?location ;
    rdfs:label ?label ;
    dct:format ?format ;
    nfo:fileName ?filename ;
    nidm_inCoordinateSpace: ?coord_space_id ;
    crypto:sha512 ?sha ;
    prov:wasGeneratedBy ?inference_id .
    OPTIONAL {?oid dc:description ?visualisation } .
}
ORDER BY ?peak_label
        """
        sd = self.graph.query(query)

        exc_sets = dict()
        if sd:
            for row in sd:
                args = row.asdict()

                coord_space = self.get_coord_spaces(row.coord_space_id)
                args['coord_space'] = coord_space
                args.pop("coord_space_id", None)

                inference = self.get_inferences(row.inference_id)
                args['inference'] = inference
                args.pop("inference_id", None)

                exc_sets[args['oid']] = ExcursionSet(**args)

        self.objects.update(exc_sets)
        return exc_sets

    def serialize(self, destination, format="mkda", overwrite=False,
                  last_used_con_id=0):
        # We need the peaks, excursion set maps and contrast maps
        self.get_peaks()
        self.get_excursion_set_maps()
        self.get_inferences()
        # self.get_contrast_maps()

        if format == "mkda":
            if not destination.endswith(".csv"):
                destination = destination + ".csv"
            csvfile = destination

            if overwrite:
                add_header = True
                open_mode = 'wb'
            else:
                add_header = False
                open_mode = 'ab'

            with open(csvfile, open_mode) as fid:
                writer = csv.writer(fid, delimiter='\t')
                if add_header:
                    writer.writerow(["9", "NaN", "NaN", "NaN", "NaN", "NaN",
                                     "NaN", "NaN", "NaN"])
                    writer.writerow(
                        ["x", "y", "z", "Study", "Contrast", "N",
                         "FixedRandom", "CoordSys", "Name"])

                # Assumed random effects
                self.FixedRandom = "random"

                # For anything that has a label
                con_ids = dict()
                con_ids[None] = last_used_con_id

                for oid, peak in self.get_peaks().items():
                    exc_set = self.objects[peak.exc_set_id]

                    if exc_set.coord_space.is_mni():
                        space = "MNI"
                    elif exc_set.coord_space.is_talairach():
                        space = "T88"
                    else:
                        raise Exception(
                            "Unrecognised space for " +
                            str(exc_set.coord_space.coordinate_system))

                    stat_map = exc_set.inference.stat_map

                    con_name = stat_map.contrast_name.replace(
                        " ", "_").replace(":", "")

                    # FIXME: need to deal with more than one group
                    self.N = stat_map.contrast_estimation.param_estimate.\
                        model_param_estimation.data.group_or_sub.num_subjects

                    if con_name in con_ids:
                        con_id = con_ids[con_name]
                    else:
                        con_id = max(con_ids.values()) + 1
                        con_ids[con_name] = con_id

                    writer.writerow([
                        peak.coordinate.coord_vector[0],
                        peak.coordinate.coord_vector[1],
                        peak.coordinate.coord_vector[2], self.study_name,
                        con_id,
                        self.N, self.FixedRandom,
                        space, con_name])

                return con_ids

        return con_ids



#!/usr/bin/env python
'''Common functions for test procedures

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>, Satrajit Ghosh
@copyright: University of Warwick 2014
'''

import os
import sys
import re
import vcr
import urllib2
import rdflib
from rdflib.graph import Graph
from rdflib.compare import *
import logging
import signal
import socket
import ssl
import glob

RELPATH = os.path.dirname(os.path.abspath(__file__))

NIDM_RESULTS_PATH = os.path.join(RELPATH, os.pardir)
NIDM_PATH = os.path.join(NIDM_RESULTS_PATH, os.pardir, os.pardir)

# Append parent script directory to path
sys.path.append(os.path.join(NIDM_PATH, "scripts"))
from nidmresults.objects.constants import \
    NIDM_SOFTWARE_VERSION, FSL_FEAT_VERSION

# Save debug info in a log file (debug.log)
logging.basicConfig(filename='debug.log', level=logging.DEBUG, filemode='w')
logger = logging.getLogger(__name__)
logger.info(' ---------- Debug log ----------')

# Complete examples (used for test queries)
import_test_filenames = set(
    glob.glob(os.path.join(NIDM_RESULTS_PATH, 'spm', '*.ttl')) +
    glob.glob(os.path.join(NIDM_RESULTS_PATH, 'spm', '*', '*.ttl')) +
    glob.glob(os.path.join(NIDM_RESULTS_PATH, 'fsl', '*', '*.ttl')) +
    glob.glob(os.path.join(NIDM_RESULTS_PATH, 'fsl', '*.ttl')) +
    glob.glob(os.path.join(NIDM_RESULTS_PATH, 'afni', '*', '*.ttl')))

# All examples
example_filenames = import_test_filenames.union(set(glob.glob(
    os.path.join(NIDM_RESULTS_PATH, 'test', 'ground_truth', '*', '*.ttl'))))

# If True turtle file will be downloaded from the prov store using the address specified in the README. 
# If False the turtle version will be retreived on the fly using the prov translator. By default set to True
# to check as README should be up to date but setting to False can be useful for local testing.
ttl_from_readme = False


def get_turtle(provn_file):
    if ttl_from_readme:
        # Get URL of turtle from README file
        readme_file = os.path.join(os.path.dirname(provn_file), 'README')
        readme_file = open(readme_file, 'r')
        readme_txt = readme_file.read()
        turtle_search = re.compile(r'.*turtle: (?P<ttl_file>.*\.ttl).*')
        extracted_data = turtle_search.search(readme_txt) 
        ttl_file_url = extracted_data.group('ttl_file');
    else:
        # Open corresponding provn file
        logger.info(' Converting '+provn_file)
        provn_file = open(provn_file, 'r')
        ex_provn = provn_file.read()

        # Convert to turtle using Prov Translator APIs
        url = "https://provenance.ecs.soton.ac.uk/validator/provapi/documents/"
        headers = { 'Content-type' : "text/provenance-notation",
                    'Accept' : "text/turtle" }
        req = urllib2.Request(url, ex_provn, headers)

        MAX_RETRY = 15
        retry = 0
        while retry <= MAX_RETRY:
            try:
                logger.info(' urllib2 open ')
                with vcr.use_cassette(
                    os.path.join(NIDM_PATH, 'vcr_cassettes/synopsis.yaml'),
                    record_mode='new_episodes',
                    match_on=['method', 'scheme', 'host', 'port', 'path',
                                  'query', 'body']):
                    response = urllib2.urlopen(req, timeout=10)
            except (socket.timeout, urllib2.URLError, ssl.SSLError):
                # On timeout retry
                retry = retry + 1 
                logger.info('Retry #'+str(retry))
                continue
            break

        if retry > MAX_RETRY:
            raise Exception("Too many retry ("+str(retry)+")")

        ttl_file_url = response.geturl()

    logger.info(' Loading turtle file '+ttl_file_url)

    return ttl_file_url

def merge_exception_dict(excep_dict, other_except_dict):
    merged_dict = dict(excep_dict.items() + other_except_dict.items())
    # When key is in both dictionaries, we need to merge the set manually
    for key in list(set(excep_dict.keys()) & set(other_except_dict.keys())):
        merged_dict[key] = excep_dict[key].union(other_except_dict[key])

    return merged_dict

def display_graph(diff_graph, prefix_msg="Difference in:"):
    found_difference = False
    for s,p,o in diff_graph.triples((None,None,None)):
            # workaround to avoid issue with "5853" being a string
            prefix, namespace, name = diff_graph.compute_qname(p)
            if p != NIDM_SOFTWARE_VERSION:
                if p != FSL_FEAT_VERSION:
                    o_name = ""
                    if isinstance(o, rdflib.URIRef):
                        unused, unused, o_name = diff_graph.compute_qname(o)
                    # Ignore prov:Location not specified explicitely
                    if o_name != 'Location':
                        found_difference = True

                        s_str = str(s)
                        if isinstance(s, rdflib.term.URIRef) \
                            and not isinstance(s, rdflib.term.BNode):
                            s_str = diff_graph.qname(s)
                        elif isinstance(s, rdflib.term.Literal):
                            s_str = s_str+" ("+str(s.datatype)+")"
                        p_str = str(p)
                        if isinstance(p, rdflib.term.URIRef) \
                            and not isinstance(p, rdflib.term.BNode):
                            p_str = diff_graph.qname(p)
                        elif isinstance(p, rdflib.term.Literal):
                            p_str = p_str+" ("+str(p.datatype)+")"                        
                        o_str = str(o)
                        if isinstance(o, rdflib.term.URIRef) \
                            and not isinstance(o, rdflib.term.BNode):
                            o_str = diff_graph.qname(o)
                        elif isinstance(o, rdflib.term.Literal):
                            o_str = o_str+" ("+str(o.datatype)+")"    
                                                    
                        logger.debug("\t"+prefix_msg+' s='+s_str+\
                                ", p="+p_str+\
                                ", o="+o_str)

    return found_difference

def compare_graphs(graph_doc1, graph_doc2):
    # Use isomorphic to ignore BNode
    iso1 = to_isomorphic(graph_doc1)
    iso2 = to_isomorphic(graph_doc2)

    found_difference = False
    if iso1 != iso2:

        in_both, in_first, in_second = graph_diff(iso1, iso2)

        # diff_graph = (in_first+in_second)
        found_difference_1 = display_graph(in_first, "\t In first: ")
        found_difference_2 = display_graph(in_second, "\t In second: ")
        found_difference = found_difference_1 or found_difference_2
        
                    # break;
    return found_difference

class TimeoutError(Exception):
    """Base class for timeout exceptions in this module."""
    pass

def _get_ttl_doc_content(doc):
    logger.info(' Opening '+doc)

    def _raise_timeout(*args):
        raise TimeoutError("end of time")

    if doc.startswith("http"):      
        # Number of retry
        MAX_RETRY = 15
        # Timeout after 5s
        TIMEOUT = 5

        retry = 0
        while retry <= MAX_RETRY:
            try:
                logger.info(' urllib2 open ')
                with vcr.use_cassette(
                        os.path.join(NIDM_PATH, 'vcr_cassettes/synopsis.yaml'),
                        record_mode='new_episodes'):
                    ttl_doc_req = urllib2.urlopen(doc, timeout=TIMEOUT)

                # There is no mechanism to handle timeout on read() in urllib2, 
                # so we need to use a timer
                signal.signal(signal.SIGALRM, _raise_timeout)                
                signal.alarm(TIMEOUT)
                logger.info(' urllib2 read ')
                doc_content = ttl_doc_req.read()
                signal.alarm(0)

            except (socket.timeout, TimeoutError, urllib2.URLError):
                # On timeout retry
                retry = retry + 1 
                logger.info(' Retry #'+str(retry))
                continue
            break

        if retry > MAX_RETRY:
            raise Exception("Too many retry ("+str(retry)+")")

    else:
        # Document locally on disk
        logger.info(' file read ')
        fid = open(doc)
        doc_content = fid.read()

    logger.info('*** ------- > read')
    return doc_content


def compare_ttl_documents(ttl_doc1, ttl_doc2):
    # Check whether most recent document is identical to current version
    doc_graph = Graph()
    doc1 = _get_ttl_doc_content(ttl_doc1)
    doc_graph.parse(data=doc1, format='turtle')
    
    same_doc_graph = Graph()
    doc2 = _get_ttl_doc_content(ttl_doc2)
    same_doc_graph.parse(data=doc2, format='turtle')

    found_difference = compare_graphs(same_doc_graph, doc_graph)

    # # Use isomorphic to ignore BNode
    # iso1 = to_isomorphic(same_doc_graph)
    # iso2 = to_isomorphic(doc_graph)

    # found_difference = False
    # if iso1 != iso2:

    #     in_both, in_first, in_second = graph_diff(iso1, iso2)

    #     # diff_graph = (in_first+in_second)
    #     found_difference_1 = display_graph(in_first, "\t In first: ")
    #     found_difference_2 = display_graph(in_second, "\t In second: ")
    #     found_difference = found_difference_1 or found_difference_2
        
    #                 # break;
    return found_difference


#!/usr/bin/env python
'''Common tests across-software for NI-DM export.
The software-specific test classes must inherit from this class.

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>, Satrajit Ghosh
@copyright: University of Warwick 2014
'''
import os
import rdflib
import re
import numpy as np
import json
import glob

import logging

# Append parent script directory to path
RELPATH = os.path.dirname(os.path.abspath(__file__))

from nidmresults.objects.constants_rdflib import *
from nidmresults.owl.owl_reader import OwlReader
from rdflib.namespace import RDF
from rdflib.graph import Graph

logger = logging.getLogger(__name__)


class TestResultDataModel(object):

    def get_readable_name(self, owl, graph, item):
        if isinstance(item, rdflib.term.Literal):
            if item.datatype:
                typeStr = graph.qname(item.datatype)
            else:
                typeStr = ''
            if typeStr:
                typeStr = "(" + typeStr + ")"
            name = "'" + item + "'" + typeStr
        elif isinstance(item, rdflib.term.URIRef):
            # Look for label
            # name = graph.label(item)
            name = owl.graph.qname(item)

            if name.startswith("ns"):
                # Namespace is not declared in the owl file, try with example
                # graph (e.g. for niiri)
                name = graph.qname(item)

            if not name.startswith("niiri"):
                m = re.search(r'\d\d\d\d$', name)
                # alphanumeric identifier
                if m is not None:
                    name += " (i.e. " + \
                        owl.get_label(item).split(":")[1] + ")"
        else:
            name = "unsupported type: " + item
        return name

    def get_alternatives(self, owl, graph, s=None, p=None, o=None):
        found = ""

        for (s_in,  p_in, o_in) in graph.triples((s,  p, o)):
            if not o:
                if not o_in.startswith(str(PROV)):
                    found += "; " + self.get_readable_name(owl, graph, o_in,)
            if not p:
                if not p_in.startswith(str(PROV)):
                    found += "; " + self.get_readable_name(owl, graph, p_in)
            if not s:
                if not s_in.startswith(str(PROV)):
                    found += "; " + self.get_readable_name(owl, graph, s_in)
        if len(found) > 200:
            found = '<many alternatives>'
        else:
            found = found[2:]
        return found

    # FIXME: Extend tests to more than one dataset (group analysis, ...)

    '''Tests based on the analysis of single-subject auditory data based on
    test01_spm_batch.m using SPM12b r5918.

    @author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>, Satrajit Ghosh
    @copyright: University of Warwick 2014
    '''

    def setUp(self, parent_gt_dir=None):
        self.my_execption = ""
        self.gt_dir = parent_gt_dir
        self.ex_graphs = dict()

    def load_graph(self, ttl_name):
        if ttl_name not in self.ex_graphs:
            ttl = ttl_name
            test_dir = os.path.dirname(ttl)

            configfile = os.path.join(test_dir, 'config.json')
            if not os.path.isfile(configfile):
                configfile = os.path.join(
                    os.path.abspath(
                        os.path.join(test_dir, os.pardir)), 'config.json')

            with open(configfile) as data_file:
                metadata = json.load(data_file)
            try:
                gt_file = [os.path.join(self.gt_dir, metadata["version"], x)
                           for x in metadata["ground_truth"]]
            except:
                # This part should be removed once SPM can modify json files
                gt_file = [os.path.join(self.gt_dir, metadata["versions"][0],
                           x)
                           for x in metadata["ground_truth"]]
            inclusive = metadata["inclusive"]
            if "version" in metadata:
                version = metadata["version"]
            else:
                # FIXME: this will be removed once spm json reader is fixed
                version = metadata["versions"][0]
            name = ttl.replace(test_dir, "")

            self.ex_graphs[ttl_name] = ExampleGraph(
                name, ttl, gt_file, inclusive, version)

        return self.ex_graphs[ttl_name]

        # Current script directory is test directory (containing test data)
        # self.test_dir = os.path.dirname(os.path.abspath(
        # inspect.getfile(inspect.currentframe())))

    def print_results(self, res):
        '''Print the results query 'res' to the console'''
        for idx, row in enumerate(res.bindings):
            rowfmt = []
            print "Item %d" % idx
            for key, val in sorted(row.items()):
                rowfmt.append('%s-->%s' % (key, val.decode()))
            print '\n'.join(rowfmt)

    def successful_retreive(self, res, info_str=""):
        '''Check if the results query 'res' contains a value for each field'''
        if not res.bindings:
            self.my_execption = info_str + """: Empty query results"""
            return False
        for idx, row in enumerate(res.bindings):
            for key, val in sorted(row.items()):
                logging.debug('%s-->%s' % (key, val.decode()))
                if not val.decode():
                    self.my_execption += "\nMissing: \t %s" % (key)
                    return False
        return True

        # if not self.successful_retreive(self.spmexport.query(query),
        #'ContrastMap and ContrastStandardErrorMap'):
        #     raise Exception(self.my_execption)

    def _replace_match(self, graph1, graph2, rdf_type):
        """
        Match classes of type 'rdf_type' across documents based on attributes.
        """

        # Retreive objects of type 'rdf_type' (e.g. prov:Entities)
        g1_terms = set(graph1.subjects(RDF.type, rdf_type))
        g2_terms = set(graph2.subjects(RDF.type, rdf_type))

        activity = False
        agent = False
        MIN_MATCHING = 2
        # For maps at least 4 attributes in common are needed for
        # matching (to deal with nifti files where format and
        # inCoordinateSpace might be the same for all)
        MIN_MAP_MATCHING = 4

        if rdf_type == PROV['Activity']:
            activity = True
        elif rdf_type == PROV['Entity']:
            # FIXME: This would be more efficiently done using the prov owl
            # file
            g1_terms = g1_terms.union(set(
                graph1.subjects(RDF.type, PROV['Bundles'])))
            g1_terms = g1_terms.union(set(
                graph1.subjects(RDF.type, PROV['Coordinate'])))
            g1_terms = g1_terms.union(set(
                graph1.subjects(RDF.type, PROV['Person'])))
            g2_terms = g2_terms.union(set(
                graph2.subjects(RDF.type, PROV['Bundles'])))
            g2_terms = g2_terms.union(
                set(graph2.subjects(RDF.type, PROV['Coordinate'])))
            g2_terms = g2_terms.union(
                set(graph2.subjects(RDF.type, PROV['Person'])))
        elif rdf_type == PROV['Agent']:
            agent = True

        for g1_term in g1_terms:
            min_matching = MIN_MATCHING

            # Look for a match in g2 corresponding to g1_term from g1
            g2_match = dict.fromkeys(g2_terms, 0)
            coord_space_found = False
            format_found = False

            for p, o in graph1.predicate_objects(g1_term):
                logging.debug("Trying to find a match for " +
                              str(graph1.qname(g1_term)) + " " +
                              str(graph1.qname(p)) + " " +
                              str(o))

                if p == NIDM_IN_COORDINATE_SPACE:
                    coord_space_found = True
                if p == DCT['format']:
                    format_found = True
                if coord_space_found and format_found:
                    min_matching = MIN_MAP_MATCHING

                # FIXME: changed "not activity" to "activity" but maybe this
                # could cause issues later on? and condition removed altogether
                # if activity or \
                #        (isinstance(o, rdflib.term.Literal) or p == RDF.type):
                for g2_term in graph2.subjects(p, o):
                    # We don't want to match agents to activities
                    if g2_term in g2_match:
                        g2_match[g2_term] += 1
                        logging.debug("Match found with " +
                                      str(graph1.qname(g2_term)))

                # If o is a string that is likely to be json check if we have
                # an equivalent json string
                same_json_array = False
                close_float = False
                if hasattr(o, 'datatype') and o.datatype == XSD['string']:
                    for g2_term, g2_o in graph2.subject_objects(p):
                        same_json_array, close_float = \
                            self._same_json_or_float(o, g2_o)
                        if same_json_array or close_float:
                            g2_match[g2_term] += 1
                            logging.debug("Match found with " +
                                          str(graph1.qname(g2_term)))

            if activity or agent:
                for s, p in graph1.subject_predicates(g1_term):
                    for g2_term in graph2.objects(s, p):
                        # We don't want to match agents to activities
                        if g2_term in g2_match:
                            g2_match[g2_term] += 1

            match_found = False
            g2_matched = set()
            for g2_term, match_index in g2_match.items():
                if max(g2_match.values()) >= min_matching:
                    if (match_index == max(g2_match.values())) \
                            and not g2_term in g2_matched:
                        # Found matching term
                        g2_matched.add(g2_term)

                        if not g1_term == g2_term:
                            g2_name = graph2.qname(g2_term).split(":")[-1]
                            new_id = g1_term + '_' + g2_name
                            logging.debug(graph1.qname(g1_term) +
                                          " is matched to " +
                                          graph2.qname(g2_term) +
                                          " and replaced by " +
                                          graph2.qname(new_id) +
                                          " (match=" + str(match_index) + ")")

                            for p, o in graph1.predicate_objects(g1_term):
                                graph1.remove((g1_term, p, o))
                                graph1.add((new_id, p, o))
                            for p, o in graph2.predicate_objects(g2_term):
                                graph2.remove((g2_term, p, o))
                                graph2.add((new_id, p, o))
                            for s, p in graph1.subject_predicates(g1_term):
                                graph1.remove((s, p, g1_term))
                                graph1.add((s, p, new_id))
                            for s, p in graph2.subject_predicates(g2_term):
                                graph2.remove((s, p, g2_term))
                                graph2.add((s, p, new_id))

                            g2_terms.remove(g2_term)
                            g2_terms.add(new_id)
                        else:
                            logging.debug(graph1.qname(g1_term) +
                                          " is matched to " +
                                          graph2.qname(g2_term) +
                                          " (match=" + str(match_index) + ")")

                        match_found = True
                        break

            if not match_found:
                logging.debug("No match found for " + graph1.qname(g1_term))

        return list([graph1, graph2])

    def _reconcile_graphs(self, graph1, graph2, recursive=10):
        """
        Reconcile: if two entities have exactly the same attributes: they
        are considered to be the same (set the same id for both)
        """

        # FIXME: reconcile entities+agents first (ignoring non attributes)
        # then reconcile activities based on everything
        # for each item select the closest match in the other graph (instead of
            # perfect match)
        # this is needed to get sensible error messages when comparing graph
        # do not do recursive anymore

        # We reconcile first entities and agents (based on data properties) and
        # then activities (based on all relations)
        graph1, graph2 = self._replace_match(graph1, graph2, PROV['Entity'])
        graph1, graph2 = self._replace_match(graph1, graph2, PROV['Agent'])
        graph1, graph2 = self._replace_match(graph1, graph2, PROV['Activity'])

        return list([graph1, graph2])

    def compare_full_graphs(self, gt_graph, other_graph, owl, include=False,
                            raise_now=False):
        ''' Compare gt_graph and other_graph '''

        # We reconcile gt_graph with other_graph
        gt_graph, other_graph = self._reconcile_graphs(gt_graph, other_graph)

        gt_m_other = gt_graph - other_graph
        if not include:
            # Check for predicates which are not in common to both graphs (XOR)
            diff_graph = gt_graph ^ other_graph
        else:
            diff_graph = gt_m_other

        # FIXME: There is probably something better than using os.path.basename
        # to remove namespaces
        exlude = list()
        missing_s = list()

        exc_wrong = ""
        exc_wrong_literal = ""
        exc_added = ""
        exc_missing = ""

        for s, p, o in diff_graph.triples((None,  None, None)):
            # If triple is found in other_graph
            if (s,  p, o) in other_graph and (not p == NIDM_SOFTWARE_VERSION):
                # If subject is *not* found in gt_graph
                if (s,  None, None) not in gt_graph:
                    if not s in exlude:
                        if not isinstance(s, rdflib.term.BNode):
                            exc_added += "\nAdded s:\t'%s' should be removed" \
                                % (self.get_readable_name(owl, other_graph, s))
                            exlude.append(s)
                # If predicate p is *not* found in gt_graph
                elif (None,  p, None) not in gt_graph:
                    if not p in exlude:
                        exc_added += "\nAdded p:\t'%s' should be removed" \
                            % (self.get_readable_name(owl, other_graph, p))
                        exlude.append(p)
                # If subject and predicate found in gt_graph, then object is
                # wrong (unless there is twice the same attribute and one value
                # is correct and the other wrong)
                # Alternatives are o_gt such as (s,p,o_gt) in gt and
                # (s,p,o_gt) *not* in other
                elif ((s,  p, None) in gt_m_other):
                    if isinstance(o, rdflib.term.Literal):
                        same_json_array = False
                        close_float = False
                        for o_gt in gt_m_other.objects(s,  p):
                            same_json_array, close_float = \
                                self._same_json_or_float(o, o_gt)

                            if same_json_array or close_float:
                                # We found a match
                                break

                        if not same_json_array and not close_float:
                            # Alternatives are o such as (s,p,o) in gt and
                            # (s,p,o) *not* in other
                            exc_wrong_literal += \
                                "\nWrong o:\t %s should be %s?" \
                                "\n\t\t ... in '%s %s %s'" \
                                % (
                                    self.get_readable_name(
                                        owl, other_graph, o),
                                    self.get_alternatives(
                                        owl, gt_m_other, s=s, p=p),
                                    self.get_readable_name(
                                        owl, other_graph, s),
                                    self.get_readable_name(
                                        owl, other_graph, p),
                                    self.get_readable_name(owl, other_graph, o)
                                )

                    elif not isinstance(o, rdflib.term.BNode):
                        exc_wrong += \
                            "\nWrong o:\t %s should be %s?" \
                            "\n\t\t ... in '%s %s %s'" \
                            % (
                                self.get_readable_name(owl, other_graph, o),
                                self.get_alternatives(owl, gt_graph, s=s, p=p),
                                self.get_readable_name(owl, other_graph, s),
                                self.get_readable_name(owl, other_graph, p),
                                self.get_readable_name(owl, other_graph, o)
                            )
                        # If object o is *not* found in gt_graph (and o is a
                        # URI, i.e. not the value of an attribute)
                elif (None,  None, o) not in gt_graph \
                        and isinstance(o, rdflib.term.URIRef):
                    if not o in exlude:
                        exc_added += "\nAdded o:\t'%s'" % (
                            self.get_readable_name(owl, other_graph, o))
                        exlude.append(o)
                # If subject and object found in gt_graph, then predicate is
                # wrong
                elif (s,  None, o) in gt_graph:
                    exc_wrong += "\nWrong p:\tBetween '%s' and '%s' " \
                                 "is '%s' (should be '%s'?)" \
                                 % (self.get_readable_name(
                                    owl, other_graph, s),
                                    self.get_readable_name(
                                        owl, other_graph, o),
                                    self.get_readable_name(
                                        owl, other_graph, p),
                                    self.get_alternatives(
                                        owl, gt_graph, s=s, o=o))
                # If predicate and object found in gt_graph, then subject is
                # wrong
                elif (None,  p, o) in gt_graph:
                    if not (s, None, None) in gt_graph:
                        if not s in exlude:
                            exc_added += "\nAdded:\t'%s'" % (s)
                            exlude.append(s)
                    else:
                        exc_wrong += "\nWrong s:\ts('%s') p('%s') o('%s')" \
                                     " not in gold std (should be s: '%s'?)" \
                                     % (self.get_readable_name(
                                        owl, other_graph, s),
                                        self.get_readable_name(
                                            owl, other_graph, p),
                                        self.get_readable_name(
                                            owl, other_graph, o),
                                        self.get_alternatives(
                                            owl, gt_graph, p=p, o=o))
                        # exc_wrong += "\nWrong s:\t'%s' \tto '%s' \tis '%s'
                        # (instead of %s)."%(os.path.basename(p),
                        # self.get_readable_name(owl, other_graph, o),
                        #  os.path.basename(s),found_subject)
                # If only subject found in gt_graph
                elif (s,  None, None) in gt_graph:
                    # gt_graph_possible_value = ""
                    # for (s_gt_graph,  p_gt_graph, o_gt_graph)
                    #  in gt_graph.triples((s,  p, None)):
                    #     gt_graph_possible_value += "; "+
                    # os.path.basename(p_gt_graph)
                    exc_added += "\nAdded:\tin '%s', '%s' \t ('%s')." \
                        % (self.get_readable_name(owl, gt_graph, s),
                           self.get_readable_name(owl, other_graph, p),
                           self.get_readable_name(owl, other_graph, o))
                else:
                    exc_added += "\nAdded:\tin '%s', '%s' \t ('%s')." \
                        % (self.get_readable_name(owl, gt_graph, s),
                           self.get_readable_name(owl, other_graph, p),
                           self.get_readable_name(owl, other_graph, o))

            # If subject and predicate are found in gt_graph
            elif (s,  p, o) in gt_graph and (not p == NIDM_SOFTWARE_VERSION):
                if include and (s,  p, None) in other_graph:
                    if isinstance(o, rdflib.term.Literal):
                        same_json_array = False
                        close_float = False

                        for o_gt in gt_graph.objects(s,  p):
                            same_json_array, close_float = \
                                self._same_json_or_float(o, o_gt)

                            if same_json_array or close_float:
                                # We found a match
                                break

                        if not same_json_array and not close_float:
                            exc_wrong_literal += \
                                "\nWrong o:\t %s should be %s?" \
                                "\n\t\t ... in '%s %s %s'" \
                                % (
                                    self.get_alternatives(
                                        owl, other_graph, s=s, p=p),
                                    self.get_readable_name(owl, gt_graph, o),
                                    self.get_readable_name(
                                        owl, other_graph, s),
                                    self.get_readable_name(
                                        owl, other_graph, p),
                                    self.get_alternatives(
                                        owl, other_graph, s=s, p=p)
                                )

                        # exc_wrong_literal += "\nWrong literal o:\t p('%s') of
                        #  s('%s') is o(%s) but should be o(%s)." % (
                        #     self.get_readable_name(owl, other_graph, p),
                        #     self.get_readable_name(owl, other_graph, s),
                        #     self.get_alternatives(owl, other_graph, s=s,
                            # p=p),
                        #     self.get_readable_name(owl, gt_graph, o))
                    elif not isinstance(o, rdflib.term.BNode):
                        exc_wrong += "\nWrong o:\ts('%s') p('%s') o('%s') not\
                        in gold std, o should be o(%s)?)" % (
                            self.get_readable_name(owl, other_graph, p),
                            self.get_readable_name(owl, other_graph, s),
                            self.get_alternatives(owl, other_graph, s=s, p=p),
                            self.get_readable_name(owl, gt_graph, o))
                        # If object o is *not* found in gt_graph (and o is a
                        # URI, i.e. not the value of an attribute)

                if not (s,  p, None) in other_graph:
                    #     if isinstance(o, rdflib.term.Literal):
                    #         if (not exc_wrong_literal):
                    #             exc_wrong_literal += "\nWrong literal o:\t
                    # p('%s') of s('%s') is ('%s') (instead o: '%s'?)"
                    # %(self.get_readable_name(owl, gt_graph, p),
                        # self.get_readable_name(owl, gt_graph, s),
                        # self.get_readable_name(owl, gt_graph, o),
                        # self.get_alternatives(owl, other_graph,s=s,p=p))
                    # else:
                    #     if (not exc_missing):
                    #         exc_missing += "\nMissing o (%s):\tp('%s')
                    # o('%s') \ton '%s'"%(type(o), self.get_readable_name(owl,
                        # gt_graph,
                    # p),self.get_readable_name(owl, gt_graph,o),
                    # self.get_readable_name(owl, gt_graph,s))
                    # If subject found in other_graph
                    if (s,  None, None) in other_graph:
                        exc_missing += "\nMissg p:\t %s is missing in %s?" \
                            "\n\t\t ... in %s %s %s'" \
                            % (
                                self.get_readable_name(owl, gt_graph, p),
                                self.get_readable_name(owl, gt_graph, s),
                                self.get_readable_name(owl, gt_graph, s),
                                self.get_readable_name(owl, gt_graph, p),
                                self.get_readable_name(owl, gt_graph, o),
                            )
                    # If subject is *not* found in other_graph
                    else:
                        if not s in missing_s:
                            if not isinstance(s, rdflib.term.BNode):
                                exc_missing += "\nMissg s:\t%s is missing " % (
                                    self.get_readable_name(owl, gt_graph, s))
                                missing_s.append(s)

        self.my_execption += exc_missing + \
            exc_added + exc_wrong + exc_wrong_literal

        if raise_now and self.my_execption:
            raise Exception(self.my_execption)

    def _same_json_or_float(self, o, o_other):
        # If string represents a json-array, then
        # compare as json data
        same_json_array = False
        if o.startswith("[") and o.endswith("]"):
            try:
                if json.loads(o) == json.loads(o_other):
                    same_json_array = True
            except ValueError:
                # Actually this string was not json
                same_json_array = False

        # If literal is a float allow for a small
        # tolerance to deal with possibly different
        # roundings
        close_float = False
        if o.datatype == XSD.float:
            if o_other.datatype == XSD.float:
                # Avoid None
                if o.value and o_other.value:
                    close_float = np.isclose(
                        o.value, o_other.value)

        return (same_json_array, close_float)


class ExampleGraph(object):
    '''Class representing a NIDM-Results examples graph to be compared to some
    ground truth graph'''

    def __init__(self, name, ttl_file, gt_ttl_files,
                 exact_comparison, version):
        self.name = name
        self.ttl_file = ttl_file

        self.gt_ttl_files = gt_ttl_files
        self.exact_comparison = exact_comparison
        self.graph = Graph()
        self.graph.parse(ttl_file, format='turtle')

        # Get NIDM-Results version for each example
        self.version = version

        if self.version != "dev":
            self.gt_ttl_files = [
                x.replace(os.path.join("nidm", "nidm"),
                          os.path.join("nidm_releases", self.version, "nidm"))
                for x in self.gt_ttl_files]

        # Owl file corresponding to version
        owl_file = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), 'owl',
            "nidm-results_" + version.replace(".", "") + ".owl")

        self.owl_file = owl_file

        owl_imports = None
        if self.version == "dev":
            owl_imports = glob.glob(
                os.path.join(os.path.dirname(owl_file),
                             os.pardir, os.pardir, "imports", '*.ttl'))
        self.owl = OwlReader(self.owl_file, owl_imports)



"""
Definition of constants.

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""

from prov.model import PROV, Namespace

NIDM = Namespace('nidm', "http://purl.org/nidash/nidm#")
NIIRI = Namespace("niiri", "http://iri.nidash.org/")
CRYPTO = Namespace(
    "crypto",
    "http://id.loc.gov/vocabulary/preservation/cryptographicHashFunctions#")
FSL = Namespace("fsl", "http://purl.org/nidash/fsl#")
SPM = Namespace("fsl", "http://purl.org/nidash/spm#")
DCT = Namespace("dct", "http://purl.org/dc/terms/")
OBO = Namespace("obo", "http://purl.obolibrary.org/obo/")
DCTYPE = Namespace("dctype", "http://purl.org/dc/dcmitype/")
NLX_OLD = Namespace("nlx_old", "http://neurolex.org/wiki/")
DC = Namespace("dc", "http://purl.org/dc/elements/1.1/")
NFO = Namespace(
    "nfo", "http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#")
SCR = Namespace("scr", "http://scicrunch.org/resolver/")
NIF = Namespace("nif", "http://uri.neuinfo.org/nif/nifstd/")

# Sciencrunch constants
SCR_FSL = SCR['SCR_002823']
SCR_SPM = SCR['SCR_007037']

# NeuroLex constants
NLX_OLD_FSL = NLX_OLD['birnlex_2067']
NIDM_FSL = NIDM['NIDM_0000167']

# NIFSTD constants
NIF_MRI = NIF['birnlex_2100']
NIF_EEG = NIF['ixl_0050003']
NIF_MEG = NIF['ixl_0050002']
NIF_PET = NIF['ixl_0050000']
NIF_SPECT = NIF['ixl_0050001']
NLX_FMRI_PROTOCOL = NIF['birnlex_2250']

# NIDM constants
FSL_FSLS_GAMMA_HRF = FSL['FSL_0000006']
NIDM_HAS_MRI_PROTOCOL = NIDM['NIDM_0000172']
NIDM_NUMBER_OF_SUBJECTS = NIDM['NIDM_0000171']
NIDM_GROUP_NAME = NIDM['NIDM_0000170']
NIDM_DATA = NIDM['NIDM_0000169']
NIDM_NIDM_RESULTS_EXPORT = NIDM['NIDM_0000166']
FSL_FEAT_VERSION = FSL['FSL_0000005']
FSL_DRIFT_CUTOFF_PERIOD = FSL['FSL_0000004']
FSL_TEMPORAL_DERIVATIVE = FSL['FSL_0000003']
FSL_GAUSSIAN_RUNNING_LINE_DRIFT_MODEL = FSL['FSL_0000002']
FSL_FSLS_GAMMA_DIFFERENCE_HRF = FSL['FSL_0000001']
NIDM_CONTRAST_EXPLAINED_MEAN_SQUARE_MAP = NIDM['NIDM_0000163']
NIDM_THRESHOLD = NIDM['NIDM_0000162']
NIDM_EQUIVALENT_THRESHOLD = NIDM['NIDM_0000161']
NIDM_P_VALUE_UNCORRECTED_CLASS = NIDM['NIDM_0000160']
NIDM_NOISE_FWHM_IN_VOXELS = NIDM['NIDM_0000159']
NIDM_NOISE_FWHM_IN_VERTICES = NIDM['NIDM_0000158']
NIDM_NOISE_FWHM_IN_UNITS = NIDM['NIDM_0000157']
NIDM_CLUSTERSIZEINRESELS = NIDM['NIDM_0000156']
NIDM_F_MRI_DESIGN = NIDM['NIDM_0000155']
NIDM_MIXED_DESIGN = NIDM['NIDM_0000154']
NIDM_EVENT_RELATED_DESIGN = NIDM['NIDM_0000153']
NIDM_BLOCK_BASED_DESIGN = NIDM['NIDM_0000152']
NIDM_SINE_BASIS_SET = NIDM['NIDM_0000151']
NIDM_LINEAR_SPLINE_BASIS_SET = NIDM['NIDM_0000150']
NIDM_SEARCH_VOLUME_IN_RESELS = NIDM['NIDM_0000149']
NIDM_RESEL_SIZE_IN_VOXELS = NIDM['NIDM_0000148']
NIDM_HEIGHT_CRITICAL_THRESHOLD_FWE_05 = NIDM['NIDM_0000147']
NIDM_HEIGHT_CRITICAL_THRESHOLD_FDR_05 = NIDM['NIDM_0000146']
NIDM_NOISE_ROUGHNESS_IN_VOXELS = NIDM['NIDM_0000145']
NIDM_RESELS_PER_VOXEL_MAP = NIDM['NIDM_0000144']
NIDM_EXPECTED_NUMBER_OF_VOXELS_PER_CLUSTER = NIDM['NIDM_0000143']
NIDM_EXPECTED_NUMBER_OF_VERTICES_PER_CLUSTER = NIDM['NIDM_0000142']
NIDM_EXPECTED_NUMBER_OF_CLUSTERS = NIDM['NIDM_0000141']
NIDM_CLUSTER_CENTER_OF_GRAVITY = NIDM['NIDM_0000140']
NIDM_COORDINATE_VECTOR_IN_VOXELS = NIDM['NIDM_0000139']
NIDM_HAS_MAXIMUM_INTENSITY_PROJECTION = NIDM['NIDM_0000138']
NIDM_SEARCH_VOLUME_IN_VERTICES = NIDM['NIDM_0000137']
NIDM_SEARCH_VOLUME_IN_UNITS = NIDM['NIDM_0000136']
NIDM_CONTRAST_VARIANCE_MAP = NIDM['NIDM_0000135']
NIDM_WITH_ESTIMATION_METHOD = NIDM['NIDM_0000134']
NIDM_VOXEL_UNITS = NIDM['NIDM_0000133']
NIDM_VOXEL_TO_WORLD_MAPPING = NIDM['NIDM_0000132']
NIDM_VOXEL_SIZE = NIDM['NIDM_0000131']
NIDM_VOXEL6CONNECTED = NIDM['NIDM_0000130']
NIDM_VOXEL26CONNECTED = NIDM['NIDM_0000129']
NIDM_VOXEL18CONNECTED = NIDM['NIDM_0000128']
NIDM_VERSION = NIDM['NIDM_0000127']
NIDM_VARIANCE_SPATIAL_MODEL = NIDM['NIDM_0000126']
NIDM_USER_SPECIFIED_THRESHOLD_TYPE = NIDM['NIDM_0000125']
NIDM_TARGET_INTENSITY = NIDM['NIDM_0000124']
NIDM_STATISTIC_TYPE = NIDM['NIDM_0000123']
NIDM_SOFTWARE_VERSION = NIDM['NIDM_0000122']
NIDM_SEARCH_VOLUME_IN_VOXELS = NIDM['NIDM_0000121']
NIDM_RANDOM_FIELD_STATIONARITY = NIDM['NIDM_0000120']
NIDM_Q_VALUE_FDR = NIDM['NIDM_0000119']
NIDM_PIXEL8CONNECTED = NIDM['NIDM_0000118']
NIDM_PIXEL4CONNECTED = NIDM['NIDM_0000117']
NIDM_P_VALUE_UNCORRECTED = NIDM['NIDM_0000116']
NIDM_P_VALUE_FWER = NIDM['NIDM_0000115']
NIDM_P_VALUE = NIDM['NIDM_0000114']
NIDM_OBJECT_MODEL = NIDM['NIDM_0000113']
NIDM_NUMBER_OF_DIMENSIONS = NIDM['NIDM_0000112']
NIDM_NUMBER_OF_CLUSTERS = NIDM['NIDM_0000111']
NIDM_GAUSSIAN_HRF = NIDM['NIDM_0000110']
NIDM_MIN_DISTANCE_BETWEEN_PEAKS = NIDM['NIDM_0000109']
NIDM_MAX_NUMBER_OF_PEAKS_PER_CLUSTER = NIDM['NIDM_0000108']
NIDM_MASKED_MEDIAN = NIDM['NIDM_0000107']
NIDM_IS_USER_DEFINED = NIDM['NIDM_0000106']
NIDM_IN_WORLD_COORDINATE_SYSTEM = NIDM['NIDM_0000105']
NIDM_IN_COORDINATE_SPACE = NIDM['NIDM_0000104']
NIDM_HAS_MAP_HEADER = NIDM['NIDM_0000103']
NIDM_HAS_HRF_BASIS = NIDM['NIDM_0000102']
NIDM_HAS_ERROR_DISTRIBUTION = NIDM['NIDM_0000101']
NIDM_HAS_ERROR_DEPENDENCE = NIDM['NIDM_0000100']
NIDM_HAS_CONNECTIVITY_CRITERION = NIDM['NIDM_0000099']
NIDM_HAS_CLUSTER_LABELS_MAP = NIDM['NIDM_0000098']
NIDM_HAS_ALTERNATIVE_HYPOTHESIS = NIDM['NIDM_0000097']
NIDM_GRAND_MEAN_SCALING = NIDM['NIDM_0000096']
NIDM_GLOBAL_NULL_DEGREE = NIDM['NIDM_0000095']
NIDM_ERROR_VARIANCE_HOMOGENEOUS = NIDM['NIDM_0000094']
NIDM_ERROR_DEGREES_OF_FREEDOM = NIDM['NIDM_0000093']
NIDM_EQUIVALENT_ZSTATISTIC = NIDM['NIDM_0000092']
NIDM_EFFECT_DEGREES_OF_FREEDOM = NIDM['NIDM_0000091']
NIDM_DIMENSIONS_IN_VOXELS = NIDM['NIDM_0000090']
NIDM_DEPENDENCE_SPATIAL_MODEL = NIDM['NIDM_0000089']
NIDM_HAS_DRIFT_MODEL = NIDM['NIDM_0000088']
NIDM_DRIFT_MODEL = NIDM['NIDM_0000087']
NIDM_COORDINATE_VECTOR = NIDM['NIDM_0000086']
NIDM_CONTRAST_NAME = NIDM['NIDM_0000085']
NIDM_CLUSTER_SIZE_IN_VOXELS = NIDM['NIDM_0000084']
NIDM_CLUSTER_SIZE_IN_VERTICES = NIDM['NIDM_0000083']
NIDM_CLUSTER_LABEL_ID = NIDM['NIDM_0000082']
NIDM_WORLD_COORDINATE_SYSTEM = NIDM['NIDM_0000081']
NIDM_VOXEL_CONNECTIVITY_CRITERION = NIDM['NIDM_0000080']
NIDM_TWO_TAILED_TEST = NIDM['NIDM_0000079']
NIDM_TALAIRACH_COORDINATE_SYSTEM = NIDM['NIDM_0000078']
NIDM_SUBJECT_COORDINATE_SYSTEM = NIDM['NIDM_0000077']
NIDM_STATISTIC_MAP = NIDM['NIDM_0000076']
NIDM_STANDARDIZED_COORDINATE_SYSTEM = NIDM['NIDM_0000075']
NIDM_SPATIALLY_REGULARIZED_MODEL = NIDM['NIDM_0000074']
NIDM_SPATIALLY_LOCAL_MODEL = NIDM['NIDM_0000073']
NIDM_SPATIALLY_GLOBAL_MODEL = NIDM['NIDM_0000072']
NIDM_SPATIAL_MODEL = NIDM['NIDM_0000071']
NIDM_SIGNIFICANT_CLUSTER = NIDM['NIDM_0000070']
NIDM_FOURIER_BASIS_SET = NIDM['NIDM_0000069']
NIDM_SEARCH_SPACE_MASK_MAP = NIDM['NIDM_0000068']
NIDM_CUSTOM_BASIS_SET = NIDM['NIDM_0000067']
NIDM_RESIDUAL_MEAN_SQUARES_MAP = NIDM['NIDM_0000066']
NIDM_POISSON_DISTRIBUTION = NIDM['NIDM_0000065']
NIDM_PIXEL_CONNECTIVITY_CRITERION = NIDM['NIDM_0000064']
NIDM_PEAK_DEFINITION_CRITERIA = NIDM['NIDM_0000063']
NIDM_PEAK = NIDM['NIDM_0000062']
NIDM_PARAMETER_ESTIMATE_MAP = NIDM['NIDM_0000061']
NIDM_ONE_TAILED_TEST = NIDM['NIDM_0000060']
NIDM_NON_PARAMETRIC_SYMMETRIC_DISTRIBUTION = NIDM['NIDM_0000059']
NIDM_NON_PARAMETRIC_DISTRIBUTION = NIDM['NIDM_0000058']
NIDM['NIDM_OBJECT_MODEL = NIDM_NIDM_0000057']
NIDM_MODEL_PARAMETERS_ESTIMATION = NIDM['NIDM_0000056']
NIDM_MNI305_COORDINATE_SYSTEM = NIDM['NIDM_0000055']
NIDM_MASK_MAP = NIDM['NIDM_0000054']
NIDM_MAP_HEADER = NIDM['NIDM_0000053']
NIDM_MAP = NIDM['NIDM_0000052']
NIDM_MNI_COORDINATE_SYSTEM = NIDM['NIDM_0000051']
NIDM_IXI549_COORDINATE_SYSTEM = NIDM['NIDM_0000050']
NIDM_INFERENCE = NIDM['NIDM_0000049']
NIDM_INDEPENDENT_ERROR = NIDM['NIDM_0000048']
NIDM_ICBM_MNI152_NON_LINEAR6TH_GENERATION_COORDINATE_SYSTEM = \
    NIDM['NIDM_0000047']
NIDM_ICBM_MNI152_NON_LINEAR2009C_SYMMETRIC_COORDINATE_SYSTEM = \
    NIDM['NIDM_0000046']
NIDM_ICBM_MNI152_NON_LINEAR2009C_ASYMMETRIC_COORDINATE_SYSTEM = \
    NIDM['NIDM_0000045']
NIDM_ICBM_MNI152_NON_LINEAR2009B_SYMMETRIC_COORDINATE_SYSTEM = \
    NIDM['NIDM_0000044']
NIDM_ICBM_MNI152_NON_LINEAR2009B_ASYMMETRIC_COORDINATE_SYSTEM = \
    NIDM['NIDM_0000043']
NIDM_ICBM_MNI152_NON_LINEAR2009A_SYMMETRIC_COORDINATE_SYSTEM = \
    NIDM['NIDM_0000042']
NIDM_ICBM_MNI152_NON_LINEAR2009A_ASYMMETRIC_COORDINATE_SYSTEM = \
    NIDM['NIDM_0000041']
NIDM_ICBM_MNI152_LINEAR_COORDINATE_SYSTEM = NIDM['NIDM_0000040']
NIDM_ICBM452_WARP5_COORDINATE_SYSTEM = NIDM['NIDM_0000039']
NIDM_ICBM452_AIR_COORDINATE_SYSTEM = NIDM['NIDM_0000038']
NIDM_HEMODYNAMIC_RESPONSE_FUNCTION_DERIVATIVE = NIDM['NIDM_0000037']
NIDM_HEMODYNAMIC_RESPONSE_FUNCTION_BASIS = NIDM['NIDM_0000036']
NIDM_HEMODYNAMIC_RESPONSE_FUNCTION = NIDM['NIDM_0000035']
NIDM_HEIGHT_THRESHOLD = NIDM['NIDM_0000034']
NIDM_GRAND_MEAN_MAP = NIDM['NIDM_0000033']
NIDM_GAUSSIAN_DISTRIBUTION = NIDM['NIDM_0000032']
NIDM_GAMMA_HRF = NIDM['NIDM_0000031']
NIDM_GAMMA_HRB = NIDM['NIDM_0000030']
NIDM_GAMMA_DIFFERENCE_HRF = NIDM['NIDM_0000029']
NIDM_FINITE_IMPULSE_RESPONSE_HRB = NIDM['NIDM_0000028']
NIDM_RESULTS = NIDM['NIDM_0000027']
NIDM_EXTENT_THRESHOLD = NIDM['NIDM_0000026']
NIDM_EXCURSION_SET_MAP = NIDM['NIDM_0000025']
NIDM_EXCHANGEABLE_ERROR = NIDM['NIDM_0000024']
NIDM_ERROR_MODEL = NIDM['NIDM_0000023']
NIDM_ERROR_DISTRIBUTION = NIDM['NIDM_0000022']
NIDM_REGRESSOR_NAMES = NIDM['NIDM_0000021']
NIDM_DISPLAY_MASK_MAP = NIDM['NIDM_0000020']
NIDM_DESIGN_MATRIX = NIDM['NIDM_0000019']
NIDM_DATA_SCALING = NIDM['NIDM_0000018']
NIDM_CUSTOM_COORDINATE_SYSTEM = NIDM['NIDM_0000017']
NIDM_COORDINATE_SPACE = NIDM['NIDM_0000016']
NIDM_COORDINATE = NIDM['NIDM_0000015']
NIDM_LEGENDRE_POLYNOMIAL_ORDER = NIDM['NIDM_0000014']
NIDM_CONTRAST_STANDARD_ERROR_MAP = NIDM['NIDM_0000013']
NIDM_CONNECTIVITY_CRITERION = NIDM['NIDM_0000012']
NIDM_CONJUNCTION_INFERENCE = NIDM['NIDM_0000011']
NIDM_HAS_FMRI_DESIGN = NIDM['NIDM_0000010']
NIDM_COLIN27_COORDINATE_SYSTEM = NIDM['NIDM_0000009']
NIDM_CLUSTER_LABELS_MAP = NIDM['NIDM_0000008']
NIDM_CLUSTER_DEFINITION_CRITERIA = NIDM['NIDM_0000007']
NIDM_CLUSTER = NIDM['NIDM_0000006']
NIDM_BINOMIAL_DISTRIBUTION = NIDM['NIDM_0000005']
NIDM_BINARY_MAP = NIDM['NIDM_0000004']
NIDM_ARBITRARILY_CORRELATED_ERROR = NIDM['NIDM_0000003']
NIDM_CONTRAST_ESTIMATION = NIDM['NIDM_0000001']
NIDM_CONTRAST_MAP = NIDM['NIDM_0000002']

STATO_OLS = OBO['STATO_0000370']
STATO_WLS = OBO['STATO_0000371']
STATO_GLS = OBO['STATO_0000372']
STATO_TSTATISTIC = OBO['STATO_0000176']
STATO_ZSTATISTIC = OBO['STATO_0000376']
STATO_FSTATISTIC = OBO['STATO_0000282']
STATO_CONTRAST_WEIGHT_MATRIX = OBO['STATO_0000323']
STATO_NORMAL_DISTRIBUTION = OBO['STATO_0000227']
STATO_GROUP = OBO['STATO_0000193']

OBO_STATISTIC = OBO['STATO_0000039']
OBO_P_VALUE_FWER = OBO['OBI_0001265']
OBO_Q_VALUE_FDR = OBO['OBI_0001442']

NIDM_INDEPEDENT_ERROR = NIDM['NIDM_0000048']
OBO_SERIALLY_CORR_COV = OBO['STATO_0000357']
OBO_COMPOUND_SYMMETRY_COV = OBO['STATO_0000362']
OBO_UNSTRUCTURED_COV = OBO['STATO_0000405']

SPATIALLY_GLOBAL = NIDM_SPATIALLY_GLOBAL_MODEL
SPATIALLY_LOCAL = NIDM_SPATIALLY_LOCAL_MODEL
SPATIALLY_REGUL = NIDM_SPATIALLY_REGULARIZED_MODEL

SPATIAL_DEPENDENCY_ENUM = (
    SPATIALLY_GLOBAL,
    SPATIALLY_LOCAL,
    SPATIALLY_REGUL
)



#!/usr/bin/env python
'''Check that a given provn file is consistent with nidm-results.owl

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2014
'''

from rdflib import RDF, term
# from rdflib.graph import Graph
import sys, os

RELPATH = os.path.dirname(os.path.abspath(__file__))

# Append parent script directory to path
sys.path.append(os.path.join(RELPATH, os.pardir, os.pardir, os.pardir, "scripts"))
from nidmresults.objects.constants import *

# ignored_attributes = set([   
#                         RDFS['label'], 
#                         RDF['type'],
#                         PROV['value'], PROV['atTime'], PROV['used'], PROV['wasAssociatedWith'],
#                         PROV['qualifiedGeneration'], PROV['wasGeneratedBy'], PROV['atLocation'], 
#                         PROV['wasDerivedFrom'], 
#                         CRYPTO['sha512']
#                         ])  

# def get_sub_class_names(my_graph):
#     sub_types = set()

#     prov_types = set([PROV['Entity'], PROV['Activity'], PROV['Agent']])
#     for prov_type in prov_types:
#         for instance_id in my_graph.subjects(RDF.type, prov_type):
#             for class_name in my_graph.objects(instance_id, RDF.type):
#                if not class_name == prov_type:
#                     sub_types.add(class_name)

#     return sub_types

# def get_class_names_in_owl(my_owl_graph):
#     # Add PROV sub-types
#     sub_types = set([PROV['Bundle'], PROV['Location'], PROV['Collection']]);

#     for class_name in my_owl_graph.subjects(RDF['type'], OWL['Class']):
#         sub_types.add(class_name)

#     return sub_types

def get_property_names_in_owl(my_owl_graph):
    properties = set();
    for class_name in my_owl_graph.subjects(RDF['type'], OWL['DatatypeProperty']):
        properties.add(class_name)
    for class_name in my_owl_graph.subjects(RDF['type'], OWL['ObjectProperty']):
        properties.add(class_name)
    return properties

# def get_attributes_from_owl(my_owl_graph):
#     attributes = dict()
#     # For each ObjectProperty found out corresponding range
#     ranges = dict()

#     restrictions = dict()

#     # Check owl restrictions on classes
#     for class_name in my_owl_graph.subjects(RDF['type'], OWL['Class']):
#         for class_restr in my_owl_graph.objects(class_name, RDFS['subClassOf']):
#             if isinstance(class_restr, term.BNode):
#                 for prop in my_owl_graph.objects(class_restr,OWL['onProperty']):
#                     attributes.setdefault(class_name, set([prop])).add(prop)
#                     for child_class in my_owl_graph.subjects(RDFS['subClassOf'], class_name): 
#                         attributes.setdefault(child_class, set([prop])).add(prop)

#     # Attributes that can be found in all classes
#     for prop,p,o in my_owl_graph.triples((None, RDF['type'], None)):
#         if o == OWL['DatatypeProperty'] or o == OWL['ObjectProperty']:

#             # Check property domain
#             for class_name in my_owl_graph.objects(prop, RDFS['domain']):
#                 # Add attribute to current class
#                 if class_name in attributes:
#                     attributes[class_name].add(prop)
#                 else:
#                     attributes[class_name] = set([prop])
                
#                 # Add attribute to children of current class
#                 for child_class in my_owl_graph.subjects(RDFS['subClassOf'], class_name):
#                     # Add attribute to current class
#                     if child_class in attributes:
#                         attributes[child_class].add(prop)
#                     else:
#                         attributes[child_class] = set([prop])
#                     class_name = child_class

#             for range_name in my_owl_graph.objects(prop, RDFS['range']):
#                 # More complex type including restrictions
#                 if isinstance(range_name, term.BNode):
#                     for restriction_node in my_owl_graph.objects(range_name, OWL['withRestrictions']):
#                         for first_restriction in my_owl_graph.objects(restriction_node, RDF['first']):
#                             xsd_restrictions = set(['minInclusive', 'minExclusive', 'maxInclusive', 'maxExclusive'])
#                             for xsd_restriction in xsd_restrictions:
#                                 for min_incl in my_owl_graph.objects(first_restriction, XSD[xsd_restriction]):
#                                     if (prop in restrictions):
#                                         if (xsd_restriction in restrictions[prop]):
#                                             restrictions[prop] = max(restrictions[prop][xsd_restriction], min_incl)
#                                         else:
#                                             restrictions[prop] = { xsd_restriction: min_incl}
#                                     else:
#                                         restrictions[prop] = { xsd_restriction: min_incl}

#                     for sub_range_name in my_owl_graph.objects(range_name, OWL['onDatatype']):
#                         range_name = sub_range_name

#                 if prop in ranges:
#                     ranges[prop].add(range_name)
#                 else:
#                     ranges[prop] = set([range_name])
#                 # FIXME: more elegant?
#                 # Add child_class to range (for ObjectProperty)
#                 for child_class in my_owl_graph.subjects(RDFS['subClassOf'], range_name):
#                     # Add range to current class
#                     if prop in ranges:
#                         ranges[prop].add(child_class)
#                     else:
#                         ranges[prop] = set([child_class])
#                     range_name = child_class
#                     for child_class in my_owl_graph.subjects(RDFS['subClassOf'], range_name):
#                         # Add attribute to current class
#                         if prop in ranges:
#                             ranges[prop].add(child_class)
#                         else:
#                             ranges[prop] = set([child_class])
#                         range_name = child_class
#                         for child_class in my_owl_graph.subjects(RDFS['subClassOf'], range_name):
#                             # Add attribute to current class
#                             if prop in ranges:
#                                 ranges[prop].add(child_class)
#                             else:
#                                 ranges[prop] = set([child_class])
#                             range_name = child_class
#                             for child_class in my_owl_graph.subjects(RDFS['subClassOf'], range_name):
#                                 # Add attribute to current class
#                                 if prop in ranges:
#                                     ranges[prop].add(child_class)
#                                 else:
#                                     ranges[prop] = set([child_class])
#                                 range_name = child_class
            
#     return list((attributes, ranges, restrictions))

# def get_owl_graph(owl_file, import_files=None):
#     # Read owl (turtle) file
#     owl_graph = Graph()
#     # This is a workaround to avoid issue with "#" in base prefix as 
#     # described in https://github.com/RDFLib/rdflib/issues/379,
#     # When the fix is introduced in rdflib these 2 lines will be replaced by:
#     # self.owl.parse(owl_file, format='turtle')
#     owl_txt = open(owl_file, 'r').read().replace("http://www.w3.org/2002/07/owl#", 
#                     "http://www.w3.org/2002/07/owl")
#     owl_graph.parse(data=owl_txt, format='turtle')

#     if import_files is not None:
#         for import_file in import_files:
#             # Read owl (turtle) file
#             import_graph = Graph()
#             import_txt = open(import_file, 'r').read()

#             # This is a workaround to avoid issue with "#" in base prefix as 
#             # described in https://github.com/RDFLib/rdflib/issues/379,
#             # When the fix is introduced in rdflib these 2 lines will be replaced by:
#             # self.owl.parse(owl_file, format='turtle')
#             import_txt = import_txt.replace("http://www.w3.org/2002/07/owl#", 
#                             "http://www.w3.org/2002/07/owl")
#             import_graph.parse(data=import_txt, format='turtle')

#             owl_graph = owl_graph + import_graph

#     return owl_graph


# def check_class_names(example_graph, example_name, class_names=None, 
#     owl_file=None, owl_imports=None):
#     my_exception = dict()
#     if not class_names:
#         if owl_file is None:
#             raise Exception("One of class_names or owl_file must be not None.")
#         else:
#             owl_graph = get_owl_graph(owl_file)
#             class_names = get_class_names_in_owl(owl_graph)

#     sub_types = get_sub_class_names(example_graph)

#     for not_recognised_sub_type in (sub_types - class_names):

#         if not not_recognised_sub_type.startswith(str(PROV)):
#             # key = example_graph.qname(not_recognised_sub_type)
#             key = "\n Unrecognised sub-type: "+example_graph.qname(not_recognised_sub_type)
#             if key in my_exception:
#                 my_exception[key].add(example_name)
#             else:
#                 my_exception[key] = set([example_name])

#     return my_exception

# def check_attributes(example_graph, example_name, owl_attributes=None, owl_ranges=None, 
#     owl_restrictions=None, owl_graph=None, owl_file=None, owl_imports=None):
#     my_exception = dict()
#     my_range_exception = dict()
#     my_restriction_exception = dict()

#     if not owl_attributes or not owl_ranges:
#         if owl_file is None:
#             raise Exception("One of class_names or owl_file must be not None.")
#         else:
#             owl_graph = get_owl_graph(owl_file, owl_imports)

#             attributes_ranges = get_attributes_from_owl(owl_graph)
#             owl_attributes = attributes_ranges[0]
#             owl_ranges = attributes_ranges[1]   
#             owl_restrictions = attributes_ranges[2]

#     # Find all attributes
#     for s,p,o in example_graph.triples((None, None, None)):
#         # To be a DataTypeProperty then o must be a literal
#         # if isinstance(o, rdflib.term.Literal):
#         if p not in ignored_attributes:
#             # *** Check domain
#             # Get all defined types of current object
#             found_attributes = False
#             class_names = ""
#             for class_name in sorted(example_graph.objects(s, RDF['type'])):
#                 attributes = owl_attributes.get(class_name)

#                 # If the current class was defined in the owl file check if current
#                 # attribute was also defined.
#                 if attributes:
#                     if p in attributes:
#                         found_attributes = True

#                 class_names += ", "+example_graph.qname(class_name)

#             # if not found_attributes:
#                 # if attributes:
#                     # if not (p in attributes):
#             if not found_attributes:
#                 key = "\n Unrecognised attribute: "+example_graph.qname(p)+\
#                             " in "+class_names[2:]
#                 if not key in my_exception:
#                     my_exception[key] = set([example_name])
#                 else:
#                     my_exception[key].add(example_name)


#             # *** Check range for ObjectProperties and DataProperties
#             if isinstance(o, term.URIRef):
#                 # An ObjectProperty can point to an instance, then we look for its type:
#                 found_range = set(example_graph.objects(o, RDF['type']))

#                 # An ObjectProperty can point to a term
#                 if not found_range:
#                     found_range = set([o])

#                     # If the term is an individual, look for its type
#                     if OWL['NamedIndividual'] in \
#                         set(owl_graph.objects(o, RDF['type'])):
#                         found_range = set(owl_graph.objects(o, RDF['type']))

#             elif isinstance(o, term.Literal):
#                 found_range = set([o.datatype])

#             correct_range = False
#             if p in owl_ranges:
#                 # If none of the class found for current ObjectProperty value is part of the range
#                 # throw an error

#                 # If the type of current value is within the authorised ranges
#                 if found_range.intersection(owl_ranges[p]):
#                     correct_range = True
#                 else:
#                     if p in owl_ranges:
#                         # A bit more complicated to deal with "positiveInteger"
#                         for owl_range in owl_ranges[p]:
#                             # FIXME: we should be able to do better than that to check that XSD['positiveInteger'] is 
#                             # in owl_ranges[p]
#                             if (XSD['positiveInteger'] == owl_range) and\
#                                  (next(iter(found_range)) == XSD['int']) and\
#                                   (o.value >= 0):
#                                     correct_range = True
#                 if not correct_range:
#                     found_range_line = ""
#                     # FIXME: This should be better handled to be able to do "if found_range"
#                     if not None in found_range:
#                         found_range_line = ', '.join(map(example_graph.qname, sorted(found_range)))
#                     owl_range_line = ""
#                     if p in owl_ranges:
#                         owl_range_line = ', '.join(map(example_graph.qname, sorted(owl_ranges[p])))

#                     key = "\n Unrecognised range: "+\
#                         found_range_line+\
#                         ' for '+example_graph.qname(p)+' should be '+\
#                         owl_range_line
#             else:
#                 # No range found for current attribute
#                 correct_range = False
#                 key = "\n No range defined for: "+\
#                         example_graph.qname(p)

#             if not correct_range:
#                 if not key in my_range_exception:
#                     my_range_exception[key] = set([example_name])
#                 else:
#                     my_range_exception[key].add(example_name)

#             if p in owl_restrictions:
#                 restrictions_ok = True
#                 if 'minInclusive' in owl_restrictions[p]:
#                     if o.value < owl_restrictions[p]['minInclusive'].value:
#                         restrictions_ok = False
#                 if 'minExclusive' in owl_restrictions[p]:
#                     if o.value <= owl_restrictions[p]['minExclusive'].value:
#                         restrictions_ok = False
#                 if 'maxInclusive' in owl_restrictions[p]:
#                     if o.value > owl_restrictions[p]['maxInclusive'].value:
#                         restrictions_ok = False
#                 if 'maxExclusive' in owl_restrictions[p]:
#                     if o.value >= owl_restrictions[p]['maxExclusive'].value:
#                         restrictions_ok = False
#                 if not restrictions_ok:
#                     key = "\n Contraints: value "+str(o.value)+\
#                         ' for '+example_graph.qname(p)+' does not observe contraints '+\
#                         ', '.join(sorted(owl_restrictions[p]))
#                     if not key in my_restriction_exception:
#                         my_restriction_exception[key] = set([example_name])
#                     else:
#                         my_restriction_exception[key].add(example_name)

#     return list((my_exception, my_range_exception, my_restriction_exception))



"""
Objects describing the Contrast Estimation activity, its inputs and outputs as
specified in NIDM-Results.

Specification: http://nidm.nidash.org/specs/nidm-results.html

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""
import numpy as np
import os
from constants import *
import nibabel as nib
from generic import *
import uuid


class Contrast(NIDMObject):

    """
    Object representing a Contrast Estimation step: including a
    ContrastEstimation activity, its inputs and outputs.
    """

    def __init__(self, contrast_num, contrast_name, weights, estimation,
                 contrast_map, stderr_or_expl_mean_sq_map, stat_map,
                 z_stat_map=None):
        super(Contrast, self).__init__()
        # FIXME: contrast_num migth only be defined in FSL if this is not
        # generic the class should be overloaded in fsl_objects
        self.contrast_num = contrast_num
        self.contrast_name = contrast_name
        self.weights = weights
        self.estimation = estimation
        self.contrast_map = contrast_map
        self.stderr_or_expl_mean_sq_map = stderr_or_expl_mean_sq_map
        self.stat_map = stat_map
        self.z_stat_map = z_stat_map

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # Create estimation activity
        self.add_object(self.estimation, nidm_version)

        # Create contrast weights
        self.add_object(self.weights, nidm_version)

        if self.contrast_map is not None:
            # Create contrast Map
            self.contrast_map.wasGeneratedBy(self.estimation)
            self.add_object(self.contrast_map, nidm_version)

        # Create Std Err. Map (T-tests) or Explained Mean Sq. Map (F-tests)
        self.stderr_or_expl_mean_sq_map.wasGeneratedBy(self.estimation)
        self.add_object(self.stderr_or_expl_mean_sq_map, nidm_version)

        # Create Statistic Map
        self.stat_map.wasGeneratedBy(self.estimation)
        self.add_object(self.stat_map, nidm_version)

        # Create Z Statistic Map
        if self.z_stat_map:
            self.z_stat_map.wasGeneratedBy(self.estimation)
            self.add_object(self.z_stat_map, nidm_version)

        return self.p


class ContrastWeights(NIDMObject):

    """
    Object representing a ContrastWeight entity.
    """

    def __init__(self, contrast_num, contrast_name, contrast_weights,
                 stat_type):
        super(ContrastWeights, self).__init__()
        self.contrast_name = contrast_name
        self.contrast_weights = contrast_weights
        self.contrast_num = contrast_num
        self.stat_type = stat_type
        self.id = NIIRI[str(uuid.uuid4())]
        self.type = STATO_CONTRAST_WEIGHT_MATRIX
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov graph.
        """
        label = "Contrast Weights: " + self.contrast_name

        if self.stat_type.lower() == "t":
            stat = STATO_TSTATISTIC
        elif self.stat_type.lower() == "z":
            stat = STATO_ZSTATISTIC
        elif self.stat_type.lower() == "f":
            stat = STATO_FSTATISTIC

        self.add_attributes((
            (PROV['type'], STATO_CONTRAST_WEIGHT_MATRIX),
            (NIDM_STATISTIC_TYPE, stat),
            (PROV['label'], label),
            (NIDM_CONTRAST_NAME, self.contrast_name),
            (PROV['value'], self.contrast_weights)))
        return self.p


class ContrastMap(NIDMObject):

    """
    Object representing a ContrastMap entity.
    """

    def __init__(self, contrast_file, contrast_num, contrast_name,
                 coord_space, export_dir, ident=None):
        super(ContrastMap, self).__init__(export_dir)
        self.num = contrast_num
        self.name = contrast_name
        if ident is None:
            self.id = NIIRI[str(uuid.uuid4())]
        else:
            self.id = ident
        filename = 'Contrast' + self.num + '.nii.gz'
        self.file = NIDMFile(self.id, contrast_file, filename, export_dir)
        self.coord_space = coord_space
        self.type = NIDM_CONTRAST_MAP
        self.prov_type = PROV['Entity']
        self.label = "Contrast Map: " + self.name

    def export(self, nidm_version):
        """
        Create prov graph.
        """
        self.add_object(self.coord_space, nidm_version)

        # Copy contrast map in export directory
        self.add_object(self.file, nidm_version)

        # Contrast Map entity
        self.add_attributes((
            (PROV['type'], NIDM_CONTRAST_MAP),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
            (NIDM_CONTRAST_NAME, self.name),
            (PROV['label'], self.label)))
        return self.p


class ContrastExplainedMeanSquareMap(NIDMObject):
    """
    Object representing a ContrastExplainedMeanSquareMap entity.
    """
    def __init__(self, stat_file, sigma_sq_file, contrast_num,
                 coord_space, export_dir):
        super(ContrastExplainedMeanSquareMap, self).__init__(export_dir)
        self.stat_file = stat_file
        self.sigma_sq_file = sigma_sq_file
        self.num = contrast_num
        self.id = NIIRI[str(uuid.uuid4())]
        self.coord_space = coord_space
        self.type = NIDM_CONTRAST_EXPLAINED_MEAN_SQUARE_MAP
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov graph.
        """
        self.add_object(self.coord_space, nidm_version)

        # Create Contrast Explained Mean Square Map as fstat<num>.nii.gz
        # multiplied by sigmasquareds.nii.gz and save it in export_dir
        fstat_img = nib.load(self.stat_file)
        fstat = fstat_img.get_data()

        sigma_sq_img = nib.load(self.sigma_sq_file)
        sigma_sq = sigma_sq_img.get_data()

        expl_mean_sq = nib.Nifti1Image(fstat*sigma_sq, fstat_img.get_qform())

        expl_mean_sq_filename = \
            "ContrastExplainedMeanSquareMap" + self.num + ".nii.gz"
        expl_mean_sq_file = os.path.join(
            self.export_dir, expl_mean_sq_filename)
        nib.save(expl_mean_sq, expl_mean_sq_file)

        self.file = NIDMFile(self.id, expl_mean_sq_file,
                             expl_mean_sq_filename, self.export_dir)
        self.add_object(self.file, nidm_version)

        # Contrast Explained Mean Square Map entity
        path, filename = os.path.split(expl_mean_sq_file)
        self.add_attributes((
            (PROV['type'], self.type),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
            (PROV['label'], "Contrast Explained Mean Square Map")))
        return self.p


class ContrastStdErrMap(NIDMObject):

    """
    Object representing a ContrastStdErrMap entity.
    """

    def __init__(self, contrast_num, filepath, is_variance, coord_space,
                 var_coord_space, export_dir):
        super(ContrastStdErrMap, self).__init__(export_dir)
        self.file = filepath
        self.id = NIIRI[str(uuid.uuid4())]
        self.is_variance = is_variance
        self.num = contrast_num
        self.coord_space = coord_space
        if is_variance:
            self.var_coord_space = var_coord_space
        self.type = NIDM_CONTRAST_STANDARD_ERROR_MAP
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov graph.
        """
        self.add_object(self.coord_space, nidm_version)

        filename = "ContrastStandardError" + self.num + ".nii.gz"
        if self.is_variance:
            self.add_object(self.var_coord_space, nidm_version)

            # Copy contrast variance map in export directory
            path, var_cope_filename = os.path.split(self.file)
            contrast_var = ContrastVariance(
                self.var_coord_space, self.file, var_cope_filename)
            self.add_object(contrast_var, nidm_version)

            # Create standard error map from contrast variance map
            var_cope_img = nib.load(self.file)
            contrast_variance = var_cope_img.get_data()

            standard_error_img = nib.Nifti1Image(np.sqrt(contrast_variance),
                                                 var_cope_img.get_qform())

            stderr_file = os.path.join(self.export_dir, filename)
            nib.save(standard_error_img, stderr_file)
            self.file = NIDMFile(
                self.id, stderr_file, filename, self.export_dir)
            self.add_object(self.file, nidm_version)

        else:
            self.file = NIDMFile(self.id, self.file, None, self.export_dir)
            self.add_object(self.file, nidm_version)

        self.add_attributes((
            (PROV['type'], self.type),
            (DCT['format'], "image/nifti"),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
            (PROV['label'], "Contrast Standard Error Map")))

        if self.is_variance:
            self.wasDerivedFrom(contrast_var)

        return self.p


class ContrastVariance(NIDMObject):
    def __init__(self, coord_space, var_file, filename):
        super(ContrastVariance, self).__init__()
        self.id = NIIRI[str(uuid.uuid4())]
        self.coord_space = coord_space
        self.type = NIDM_CONTRAST_VARIANCE_MAP
        self.file = NIDMFile(self.id, var_file)
        self.filename = filename
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        # FIXME: Use ContrastVariance.nii.gz?
        # var_cope_file = os.path.join(self.export_dir, var_cope_filename)
        # var_cope_original_filename, var_cope_filename =
        # self.copy_file(var_cope_original_file, var_cope_file)

        # Contrast Variance Map entity
        # self.provBundle.entity('niiri:'+'contrast_variance_map_id_'+
        # contrast_num,
        # other_attributes=(

        self.add_object(self.file, nidm_version)

        self.add_attributes([(PROV['type'], NIDM_CONTRAST_VARIANCE_MAP)])

        return self.p


class StatisticMap(NIDMObject):

    """
    Object representing a StatisticMap entity.
    """

    def __init__(self, location, stat_type, contrast_num, contrast_name, dof,
                 coord_space, export_dir=None, label=None, oid=None,
                 format="image/nifti", effdof=None, filename=None, sha=None,
                 contrast_estimation=None):
        super(StatisticMap, self).__init__(export_dir, oid=oid)
        self.num = contrast_num
        self.contrast_name = contrast_name
        self.stat_type = stat_type
        if self.stat_type.lower() == "t":
            self.stat = STATO_TSTATISTIC
        elif self.stat_type.lower() == "z":
            self.stat = STATO_ZSTATISTIC
        elif self.stat_type.lower() == "f":
            self.stat = STATO_FSTATISTIC
        # FIXME use new 'preferred mathematical notation from stato'
        if self.num is not None:
            filename = self.stat_type.upper() + \
                'Statistic' + self.num + '.nii.gz'
        self.file = NIDMFile(self.id, location, filename, export_dir, sha=sha)
        self.coord_space = coord_space
        self.dof = dof
        self.type = NIDM_STATISTIC_MAP
        self.prov_type = PROV['Entity']
        if label is not None:
            self.label = label
        else:
            self.label = "Statistic Map: " + self.contrast_name
            if self.stat_type == 'Z':
                self.label = self.stat_type + '-' + self.label

        self.format = format
        if effdof is None:
            # FIXME: this should not be 1 for F-test
            effdof = 1.0

        self.effdof = effdof

        # Only used when reading (so far)
        self.contrast_estimation = contrast_estimation

    def __str__(self):
        return '%s\t%s' % (self.label, self.file)

    def export(self, nidm_version):
        """
        Create prov graph.
        """
        self.add_object(self.coord_space, nidm_version)

        # Copy Statistical map in export directory
        self.add_object(self.file, nidm_version)

        attributes = [(PROV['type'], NIDM_STATISTIC_MAP),
                      (DCT['format'], self.format),
                      (PROV['label'], self.label),
                      (NIDM_STATISTIC_TYPE, self.stat),
                      (NIDM_CONTRAST_NAME, self.contrast_name),
                      (NIDM_IN_COORDINATE_SPACE, self.coord_space.id)]

        if not self.stat_type == 'Z':
            attributes.insert(0, (NIDM_ERROR_DEGREES_OF_FREEDOM, self.dof))
            attributes.insert(0, (NIDM_EFFECT_DEGREES_OF_FREEDOM, self.effdof))
        else:
            # For Z-Statistic error dof is infinity and effect dof is 1
            attributes.insert(0, (NIDM_ERROR_DEGREES_OF_FREEDOM, float("inf")))
            attributes.insert(0, (NIDM_EFFECT_DEGREES_OF_FREEDOM, self.effdof))

        # Create "Statistic Map" entity
        # FIXME: Deal with other than t-contrast maps: dof + statisticType
        self.add_attributes(attributes)
        return self.p


class ContrastEstimation(NIDMObject):

    """
    Object representing a ContrastEstimation entity.
    """

    def __init__(self, contrast_num, contrast_name=None, label=None,
                 param_estimate=None, oid=None):
        super(ContrastEstimation, self).__init__(oid=oid)
        self.num = contrast_num
        self.type = NIDM_CONTRAST_ESTIMATION
        self.prov_type = PROV['Activity']
        if label is not None:
            self.label = label
        else:
            self.label = "Contrast estimation: " + contrast_name
        # Only used when reading (so far)
        self.param_estimate = param_estimate

    def export(self, nidm_version):
        """
        Create prov graph.
        """
        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['label'], self.label)))

        return self.p



"""
Objects describing the Inference activity, its inputs and outputs as specified
in NIDM-Results.

Specification: http://nidm.nidash.org/specs/nidm-results.html

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""
from constants import *
from generic import *
import uuid
from math import erf, sqrt


class Inference(NIDMObject):

    """
    Object representing an Inference step: including an Inference activity, its
    inputs and outputs.
    """

    def __init__(
            self, inference, height_thresh, extent_thresh,
            peak_criteria, cluster_criteria, disp_mask, excursion_set,
            clusters, search_space, software_id):
        super(Inference, self).__init__()
        self.excursion_set = excursion_set
        self.inference_act = inference
        self.height_thresh = height_thresh
        self.extent_thresh = extent_thresh
        self.clusters = clusters
        self.software_id = software_id
        self.peak_criteria = peak_criteria
        self.cluster_criteria = cluster_criteria
        self.disp_mask = disp_mask
        self.search_space = search_space

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # Excursion set
        self.excursion_set.wasGeneratedBy(self.inference_act)
        self.add_object(self.excursion_set, nidm_version)

        # Height threshold
        self.add_object(self.height_thresh, nidm_version)

        # Extent threshold
        self.add_object(self.extent_thresh, nidm_version)

        # Display Mask (potentially more than 1)
        if self.disp_mask:
            for mask in self.disp_mask:
                self.inference_act.used(mask)
                self.add_object(mask, nidm_version)

        # Search Space
        self.search_space.wasGeneratedBy(self.inference_act)
        self.add_object(self.search_space, nidm_version)

        # Peak Definition
        if self.peak_criteria:
            self.inference_act.used(self.peak_criteria)
            self.add_object(self.peak_criteria, nidm_version)

        # Cluster Definition
        if self.cluster_criteria:
            self.inference_act.used(self.cluster_criteria)
            self.add_object(self.cluster_criteria, nidm_version)

        if self.clusters:
            # Clusters and peaks
            for cluster in self.clusters:
                cluster.wasDerivedFrom(self.excursion_set)
                self.add_object(cluster, nidm_version)

        # Inference activity
        self.inference_act.wasAssociatedWith(self.software_id)
        self.inference_act.used(self.height_thresh)
        self.inference_act.used(self.extent_thresh)
        self.add_object(self.inference_act, nidm_version)

        # self.p.wasGeneratedBy(NIIRI['search_space_id'],
            # self.inference_act.id)
        # self.p.used(self.inference_act.id, NIIRI['z_statistic_map_id_'+
            # contrast_num])
        # self.p.used(self.inference_act.id, NIIRI['mask_id_1'])

        return self.p


class InferenceActivity(NIDMObject):

    """
    Object representing an Inference activity.
    """

    def __init__(self, contrast_num=None, contrast_name=None, oid=None,
                 tail=None, label=None, stat_map=None):
        super(InferenceActivity, self).__init__(oid=oid)
        self.id = NIIRI[str(uuid.uuid4())]
        self.contrast_name = contrast_name
        self.type = NIDM_INFERENCE
        self.prov_type = PROV['Activity']
        if tail is None:
            tail = NIDM_ONE_TAILED_TEST
        self.tail = tail
        if label is None:
            label = "Inference: " + self.contrast_name
        self.label = label
        # FIXME: not used yet for export (only for reading)
        self.stat_map = stat_map

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """

        # In FSL we have a single thresholding (extent, height) applied to all
        # contrasts
        # FIXME: Deal with two-tailed inference?
        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['label'], self.label),
            (NIDM_HAS_ALTERNATIVE_HYPOTHESIS, self.tail)))

        return self.p


class ExcursionSet(NIDMObject):

    """
    Object representing a ExcursionSet entity.
    """

    def __init__(self, location, coord_space, visualisation=None,
                 export_dir=None, oid=None, format=None, label=None,
                 sha=None, filename=None, inference=None, suffix=''):
        super(ExcursionSet, self).__init__(export_dir, oid)
        # Excursion set is going to be copied over to export_dir folder
        if export_dir is not None:
            filename = 'ExcursionSet' + suffix + '.nii.gz'
        else:
            filename = location
        self.filename = filename
        self.file = NIDMFile(self.id, location, filename, export_dir, sha)
        self.type = NIDM_EXCURSION_SET_MAP
        self.prov_type = PROV['Entity']
        if visualisation is not None:
            visu_filename = 'ExcursionSet' + suffix + '.png'
            self.visu = Image(export_dir, visualisation, visu_filename)
        if label is None:
            label = "Excursion Set Map"
        self.label = label
        self.coord_space = coord_space
        # FIXME Not used for export yet (only for reading)
        self.inference = inference

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_object(self.coord_space, nidm_version)
        self.add_object(self.visu, nidm_version)

        # Copy "Excursion set map" file in export directory
        self.add_object(self.file, nidm_version)

        # Create "Excursion set" entity
        self.add_attributes((
            (PROV['type'], self.type),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
            (PROV['label'], self.label),
            (DC['description'], self.visu.id)
        ))

        return self.p


class HeightThreshold(NIDMObject):

    """
    Object representing a HeightThreshold entity.
    """

    def __init__(self, stat_threshold=None, p_corr_threshold=None,
                 p_uncorr_threshold=None):
        super(HeightThreshold, self).__init__()
        if not stat_threshold and not p_corr_threshold and \
           not p_uncorr_threshold:
            raise Exception('No threshold defined')

        self.stat_threshold = stat_threshold
        self.p_corr_threshold = p_corr_threshold
        self.p_uncorr_threshold = p_uncorr_threshold
        self.id = NIIRI[str(uuid.uuid4())]
        self.type = NIDM_HEIGHT_THRESHOLD
        self.prov_type = PROV['Entity']

    def export(self, version):
        """
        Create prov entities and activities.
        """
        thresh_desc = ""
        if self.stat_threshold is not None:
            thresh_desc = "Z>" + str(self.stat_threshold)
            if version['num'] == "1.0.0":
                user_threshold_type = "Z-Statistic"
            else:
                threshold_type = OBO_STATISTIC
                value = self.stat_threshold
        elif self.p_uncorr_threshold is not None:
            thresh_desc = "p<" + \
                str(self.p_uncorr_threshold) + " (uncorrected)"
            if version['num'] == "1.0.0":
                user_threshold_type = "p-value uncorrected"
            else:
                threshold_type = NIDM_P_VALUE_UNCORRECTED_CLASS
                value = self.p_uncorr_threshold
        elif self.p_corr_threshold is not None:
            thresh_desc = "p<" + str(self.p_corr_threshold) + " (FWE)"
            if version['num'] == "1.0.0":
                user_threshold_type = "p-value FWE"
            else:
                threshold_type = OBO_P_VALUE_FWER
                value = self.p_corr_threshold

        atts = [
            (PROV['type'], self.type),
            (PROV['label'], "Height Threshold: " + thresh_desc),
        ]

        if version['num'] == "1.0.0":
            atts += [
                (NIDM_USER_SPECIFIED_THRESHOLD_TYPE, user_threshold_type),
                (PROV['value'], self.stat_threshold),
                (NIDM_P_VALUE_UNCORRECTED, self.p_uncorr_threshold),
                (NIDM_P_VALUE_FWER, self.p_corr_threshold)
                ]
        else:
            atts += [
                (PROV['type'], threshold_type),
                (PROV['value'], value)
                ]

        self.add_attributes([(k, v) for k, v in atts if v is not None])
        return self.p


class ExtentThreshold(NIDMObject):

    """
    Object representing an ExtentThreshold entity.
    """

    def __init__(self, extent=None, p_corr=None, p_uncorr=None):
        super(ExtentThreshold, self).__init__()
        self.extent = extent
        self.p_corr = p_corr
        self.p_uncorr = p_uncorr
        self.id = NIIRI[str(uuid.uuid4())]
        self.type = NIDM_EXTENT_THRESHOLD
        self.prov_type = PROV['Entity']

    def export(self, version):
        """
        Create prov entities and activities.
        """
        atts = [
            (PROV['type'], self.type),
        ]

        thresh_desc = ""
        if self.extent is not None:
            thresh_desc = "k>" + str(self.extent)
            # NIDM-Results 1.0.0
            user_threshold_type = "Cluster-size in voxels"
            # NIDM-Results > 1.0.0
            threshold_type = OBO_STATISTIC
        elif not self.p_uncorr is None:
            thresh_desc = "p<" + str(self.p_uncorr) + " (uncorrected)"
            # NIDM-Results 1.0.0
            user_threshold_type = "p-value uncorrected"
            # NIDM-Results > 1.0.0
            threshold_type = NIDM_P_VALUE_UNCORRECTED_CLASS
            value = self.p_uncorr
        elif not self.p_corr is None:
            thresh_desc = "p<" + str(self.p_corr) + " (FWE)"
            # NIDM-Results 1.0.0
            user_threshold_type = "p-value FWE"
            # NIDM-Results > 1.0.0
            threshold_type = OBO_P_VALUE_FWER
            value = self.p_corr
        else:
            thresh_desc = "k>=0"
            self.extent = 0
            if version['num'] == "1.0.0":
                self.p_uncorr = 1.0
                self.p_corr = 1.0
                user_threshold_type = None
            else:
                threshold_type = OBO_STATISTIC

        atts += [
            (PROV['label'], "Extent Threshold: " + thresh_desc)
        ]

        if self.extent is not None:
            atts += [
                (NIDM_CLUSTER_SIZE_IN_VOXELS, self.extent),
            ]

        if version['num'] == "1.0.0":
            atts += [
                (NIDM_USER_SPECIFIED_THRESHOLD_TYPE, user_threshold_type),
                (NIDM_P_VALUE_UNCORRECTED, self.p_uncorr),
                (NIDM_P_VALUE_FWER, self.p_corr)
            ]
        else:
            atts += [
                (PROV['type'], threshold_type)
            ]
            if self.extent is None:
                atts += [
                    (PROV['value'], value)
                ]

        self.add_attributes([(k, v) for k, v in atts if v is not None])
        return self.p


class Cluster(NIDMObject):

    """
    Object representing a Cluster entity.
    """

    def __init__(self, cluster_num, size, pFWER, peaks,
                 x=None, y=None, z=None, x_std=None, y_std=None, z_std=None,
                 suffix=''):
        super(Cluster, self).__init__()
        self.num = cluster_num
        self.id = NIIRI[str(uuid.uuid4())]
        self.cog = CenterOfGravity(
            cluster_num, x=x, y=y, z=z, x_std=x_std, y_std=y_std, z_std=z_std)
        self.peaks = peaks
        self.size = size
        self.pFWER = pFWER
        self.type = NIDM_SIGNIFICANT_CLUSTER
        self.prov_type = PROV['Entity']
        cluster_naming = "Supra-Threshold Cluster"
        self.label = "%s %04d" % (cluster_naming, self.num)

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        for peak in self.peaks:
            peak.wasDerivedFrom(self)
            self.add_object(peak, nidm_version)

        self.cog.wasDerivedFrom(self)
        self.add_object(self.cog, nidm_version)

        if nidm_version['num'] in ["1.0.0", "1.1.0"]:
            cluster_naming = "Significant Cluster"
        else:
            cluster_naming = "Supra-Threshold Cluster"

        # FIXME deal with multiple contrasts
        self.add_attributes((
            (PROV['type'], NIDM_SIGNIFICANT_CLUSTER),
            (PROV['label'], "%s %04d" % (cluster_naming, self.num)),
            (NIDM_CLUSTER_LABEL_ID, self.num),
            (NIDM_CLUSTER_SIZE_IN_VOXELS, self.size),
            (NIDM_P_VALUE_FWER, self.pFWER)))

        return self.p


class DisplayMaskMap(NIDMObject):

    """
    Object representing a DisplayMaskMap entity.
    """
    def __init__(self, contrast_num, mask_file, mask_num, coord_space,
                 export_dir):
        super(DisplayMaskMap, self).__init__(export_dir)
        self.id = NIIRI[str(uuid.uuid4())]
        self.mask_num = mask_num
        filename = 'DisplayMask' + str(self.mask_num) + '.nii.gz'
        self.file = NIDMFile(self.id, mask_file, filename, export_dir)
        self.coord_space = coord_space
        self.type = NIDM_DISPLAY_MASK_MAP
        self.prov_type = PROV['Entity']
        self.label = "Display Mask Map " + str(mask_num)

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # Create coordinate space entity
        self.add_object(self.coord_space, nidm_version)

        # Create "Display Mask Map" entity
        self.add_object(self.file, nidm_version)

        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['label'], self.label),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id)
            ))

        return self.p


class PeakCriteria(NIDMObject):

    """
    Object representing a PeakCriteria entity.
    """

    def __init__(self, contrast_num, num_peak, peak_dist):
        super(PeakCriteria, self).__init__()
        self.id = NIIRI[str(uuid.uuid4())]
        self.num_peak = num_peak
        self.peak_dist = peak_dist
        self.type = NIDM_PEAK_DEFINITION_CRITERIA
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        num_peak = ()
        if self.num_peak:
            num_peak = (NIDM_MAXNUMBEROFPEAKSPERCLUSTER, self.num_peak)

        # Create "Peak definition criteria" entity
        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['label'], "Peak Definition Criteria"),
            (NIDM_MIN_DISTANCE_BETWEEN_PEAKS, self.peak_dist)
            ) + num_peak)

        return self.p


class ClusterCriteria(NIDMObject):

    """
    Object representing a ClusterCriteria entity.
    """

    def __init__(self, contrast_num, connectivity):
        super(ClusterCriteria, self).__init__()
        self.id = NIIRI[str(uuid.uuid4())]
        self.connectivity = connectivity
        self.type = NIDM_CLUSTER_DEFINITION_CRITERIA
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # Create "Cluster definition criteria" entity
        if self.connectivity == 6:
            voxel_conn = NIDM_VOXEL6CONNECTED
        elif self.connectivity == 18:
            voxel_conn = NIDM_VOXEL18CONNECTED
        elif self.connectivity == 26:
            voxel_conn = NIDM_VOXEL26CONNECTED

        label = "Cluster Connectivity Criterion: " + str(self.connectivity)

        # FIXME if connectivity is missing
        if self.connectivity is not None:
            atts = (
                (PROV['type'], self.type),
                (PROV['label'], label),
                (NIDM_HAS_CONNECTIVITY_CRITERION, voxel_conn))
        else:
            atts = (
                (PROV['type'], NIDM_CLUSTER_DEFINITION_CRITERIA),
                (PROV['label'], label))

        self.add_attributes(atts)

        return self.p


class CenterOfGravity(NIDMObject):

    """
    Object representing a CenterOfGravity entity.
    """

    def __init__(self, cluster_num, x=None, y=None, z=None, x_std=None,
                 y_std=None, z_std=None):
        super(CenterOfGravity, self).__init__()
        self.cluster_num = cluster_num
        self.id = NIIRI[str(uuid.uuid4())]
        self.coordinate = Coordinate("%04d" % cluster_num, x=x, y=y, z=z,
                                     x_std=x_std, y_std=y_std, z_std=z_std)
        self.type = NIDM_CLUSTER_CENTER_OF_GRAVITY
        self.prov_type = PROV['Entity']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_object(self.coordinate, nidm_version)

        label = "Center of gravity " + str(self.cluster_num)

        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['label'], label),
            (PROV['location'], self.coordinate.id)))

        return self.p


class SearchSpace(NIDMObject):

    """
    Object representing a SearchSpace entity.
    """

    def __init__(self, search_space_file, vol_in_voxels, vol_in_units,
                 vol_in_resels, resel_size_in_voxels, dlh,
                 random_field_stationarity, noise_fwhm_in_voxels,
                 noise_fwhm_in_units, coord_space, export_dir):
        super(SearchSpace, self).__init__(export_dir)
        self.id = NIIRI[str(uuid.uuid4())]
        filename = 'SearchSpaceMask.nii.gz'
        self.file = NIDMFile(self.id, search_space_file, filename, export_dir)
        self.coord_space = coord_space
        self.resel_size_in_voxels = resel_size_in_voxels
        self.dlh = dlh
        self.search_volume_in_voxels = vol_in_voxels
        self.search_volume_in_units = vol_in_units
        self.search_volume_in_resels = vol_in_resels
        self.rf_stationarity = random_field_stationarity
        self.noise_fwhm_in_voxels = noise_fwhm_in_voxels
        self.noise_fwhm_in_units = noise_fwhm_in_units
        self.type = NIDM_SEARCH_SPACE_MASK_MAP
        self.prov_type = PROV['Entity']
        self.label = "Search Space Mask Map"

    # Generate prov for search space entity generated by the inference activity
    def export(self, version):
        """
        Create prov entities and activities.
        """
        self.add_object(self.coord_space, version)

        # Copy "Mask map" in export directory
        self.add_object(self.file, version)

        atts = (
            (PROV['label'], self.label),
            (PROV['type'], NIDM_SEARCH_SPACE_MASK_MAP),
            (NIDM_RANDOM_FIELD_STATIONARITY, self.rf_stationarity),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
            (NIDM_SEARCH_VOLUME_IN_VOXELS, self.search_volume_in_voxels),
            (NIDM_SEARCH_VOLUME_IN_UNITS, self.search_volume_in_units),
            (NIDM_SEARCH_VOLUME_IN_RESELS, self.search_volume_in_resels),
            (NIDM_RESEL_SIZE_IN_VOXELS, self.resel_size_in_voxels),
            (NIDM_NOISE_ROUGHNESS_IN_VOXELS, self.dlh))

        # Noise FWHM was introduced in NIDM-Results 1.1.0
        if self.noise_fwhm_in_voxels is not None:
            if (version['major'] > 1) or \
               (version['major'] >= 1 and
                    (version['minor'] > 0 or version['revision'] > 0)):
                atts = atts + (
                    (NIDM_NOISE_FWHM_IN_VOXELS, self.noise_fwhm_in_voxels),
                    (NIDM_NOISE_FWHM_IN_UNITS, self.noise_fwhm_in_units))

        # Create "Search Space Mask map" entity
        self.add_attributes(atts)

        return self.p


class Coordinate(NIDMObject):

    """
    Object representing a Coordinate entity.
    """

    def __init__(self, label_id, coord_vector=None, coord_vector_std=None,
                 x=None, y=None, z=None, x_std=None, y_std=None, z_std=None,
                 label=None):
        super(Coordinate, self).__init__()
        # FIXME: coordiinate_id should not be determined externally
        self.id = NIIRI[str(uuid.uuid4())]
        self.label_id = label_id
        if x is not None and y is not None and z is not None:
            self.coord_vector = [x, y, z]
        else:
            if isinstance(coord_vector, rdflib.term.Literal):
                coord_vector = json.loads(coord_vector)
            self.coord_vector = coord_vector
        if x_std is not None and y_std is not None and z_std is not None:
            self.coord_vector_std = [x_std, y_std, z_std]
        else:
            self.coord_vector_std = coord_vector_std
        self.type = NIDM_COORDINATE
        self.prov_type = PROV['Entity']
        if label is not None:
            self.label = label
        else:
            self.label = "Coordinate " + self.label_id

    def __str__(self):
        return '%s\t%s' % (self.label, self.coord_vector)

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        # We can not have this in the dictionnary because we want to keep the
        # duplicate prov:type attribute
        type_label = [  # (PROV['type'],PROV['Location']),
            (PROV['type'], NIDM_COORDINATE),
            (PROV['label'], self.label)]

        coordinate = {
            NIDM_COORDINATE_VECTOR_IN_VOXELS: json.dumps(self.coord_vector),
            NIDM_COORDINATE_VECTOR: json.dumps(self.coord_vector_std),
        }

        self.add_attributes(
            type_label +
            list(dict((k, v) for k, v in coordinate.iteritems()
                      if not v is None).items()))

        return self.p


class Peak(NIDMObject):

    """
    Object representing a Peak entity.
    """

    def __init__(self, equiv_z, p_unc=None, p_fwer=None, label=None,
                 coord_label=None, exc_set_id=None, oid=None, suffix='',
                 *args, **kwargs):
        super(Peak, self).__init__(oid)
        # FIXME: Currently assumes less than 10 clusters per contrast
        # cluster_num = cluster_index
        # FIXME: Currently assumes less than 100 peaks
        if oid is not None:
            self.label = label
            peak_unique_id = label[5:]
            peak_index = peak_unique_id
            # cluster_index, peak_index = peak_unique_id.split("_")
        else:
            peak_unique_id = suffix
            self.label = "Peak " + peak_unique_id
        self.equiv_z = equiv_z
        self.p_unc = p_unc
        self.p_fwer = p_fwer
        self.coordinate = Coordinate(
            str(peak_unique_id), label=coord_label, **kwargs)
        self.type = NIDM_PEAK
        self.prov_type = PROV['Entity']
        # self.cluster = cluster_id
        self.exc_set_id = exc_set_id

    def __str__(self):
        return '%s \tz=%.2f \tp=%.2e (unc.) \t%s' % (
            self.label, self.equiv_z, self.p_unc, str(self.coordinate))

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_object(self.coordinate, nidm_version)

        if self.p_unc is None:
            norm_cdf_z = (1.0 + erf(self.equiv_z / sqrt(2.0))) / 2.0
            self.p_unc = 1 - norm_cdf_z

        self.add_attributes([
            (PROV['type'], self.type),
            (PROV['label'], self.label),
            (NIDM_EQUIVALENT_ZSTATISTIC, self.equiv_z),
            (NIDM_P_VALUE_UNCORRECTED, self.p_unc),
            (PROV['location'], self.coordinate.id)])

        return self.p
